import asyncio
import logging
from time import perf_counter
from typing import Optional, Any, List

from telegram import (
    Bot,
    InlineKeyboardMarkup,
    InlineKeyboardButton,
    Update,
    BotCommand,
    User,
    Chat,
)
from telegram.error import BadRequest
from telegram.ext import (
    CallbackQueryHandler,
    ContextTypes,
    CommandHandler,
    MessageHandler,
    BaseHandler,
)
from telegram.ext.filters import ChatType, StatusUpdate

from suppgram.analytics import Reporter
from suppgram.backend import Backend
from suppgram.entities import (
    Conversation,
    NewUnassignedMessageFromCustomerEvent,
    ConversationEvent,
    ConversationState,
    Tag,
    ConversationTagEvent,
    TagEvent,
)
from suppgram.errors import AgentNotFound, PermissionDenied, TagAlreadyExists
from suppgram.frontend import (
    ManagerFrontend,
)
from suppgram.frontends.telegram.app_manager import TelegramAppManager
from suppgram.frontends.telegram.callback_actions import CallbackActionKind
from suppgram.frontends.telegram.helper import TelegramHelper
from suppgram.frontends.telegram.helpers import (
    send_text_answer,
    arrange_buttons,
    encode_callback_data,
    decode_callback_data,
)
from suppgram.frontends.telegram.identification import (
    make_agent_identification,
    make_agent_diff,
)
from suppgram.frontends.telegram.storage import (
    TelegramStorage,
    TelegramChatRole,
    TelegramMessageKind,
    TelegramChat,
    TelegramMessage,
)
from suppgram.helpers import flat_gather
from suppgram.texts.interface import TextProvider

logger = logging.getLogger(__name__)


class TelegramManagerFrontend(ManagerFrontend):
    _SEND_NEW_CONVERSATIONS_COMMAND = "send_new_conversations"
    _AGENTS_COMMAND = "agents"
    _CREATE_CONVERSATION_TAG_COMMAND = "create_tag"
    _REPORT_COMMAND = "report"

    def __init__(
        self,
        token: str,
        app_manager: TelegramAppManager,
        backend: Backend,
        helper: TelegramHelper,
        reporter: Reporter,
        storage: TelegramStorage,
        texts: TextProvider,
    ) -> None:
        self._backend = backend
        self._helper = helper
        self._reporter = reporter
        self._storage = storage
        self._texts = texts
        self._telegram_app = app_manager.get_app(token)
        self._telegram_bot: Bot = self._telegram_app.bot

        backend.on_new_conversation.add_handler(self._handle_new_conversation_event)
        backend.on_new_unassigned_message_from_customer.add_handler(
            self._handle_new_unassigned_message_from_customer_event
        )
        backend.on_conversation_assignment.add_handler(self._handle_conversation_assignment_event)
        backend.on_conversation_resolution.add_handler(self._handle_conversation_resolution_event)
        backend.on_conversation_tag_added.add_handler(self._handle_conversation_tags_event)
        backend.on_conversation_tag_removed.add_handler(self._handle_conversation_tags_event)
        backend.on_tag_created.add_handler(self._handle_tag_event)
        self._telegram_app.add_handlers(self._make_handlers())

    def _make_handlers(self) -> List[BaseHandler]:
        return [
            CallbackQueryHandler(self._handle_callback_query),
            CommandHandler("start", self._handle_start_command, filters=ChatType.PRIVATE),
            CommandHandler(
                self._AGENTS_COMMAND,
                self._handle_agents_command,
                filters=ChatType.GROUP,
            ),
            CommandHandler(
                self._SEND_NEW_CONVERSATIONS_COMMAND,
                self._handle_send_new_conversations_command,
                filters=ChatType.GROUP,
            ),
            CommandHandler(
                self._CREATE_CONVERSATION_TAG_COMMAND,
                self._handle_create_conversation_tag_command,
                filters=ChatType.GROUP | ChatType.PRIVATE,
            ),
            CommandHandler(
                self._REPORT_COMMAND,
                self._handle_report_command,
                filters=ChatType.GROUP | ChatType.PRIVATE,
            ),
            MessageHandler(StatusUpdate.NEW_CHAT_MEMBERS, self._handle_new_chat_members),
            MessageHandler(StatusUpdate.LEFT_CHAT_MEMBER, self._handle_left_chat_member),
        ]

    async def initialize(self) -> None:
        await super().initialize()
        await self._telegram_app.initialize()
        await self._set_commands()

    async def _set_commands(self) -> None:
        await self._telegram_bot.set_my_commands(
            [
                BotCommand(
                    self._AGENTS_COMMAND,
                    self._texts.telegram_agents_command_description,
                ),
                BotCommand(
                    self._SEND_NEW_CONVERSATIONS_COMMAND,
                    self._texts.telegram_send_new_conversations_command_description,
                ),
                BotCommand(
                    self._CREATE_CONVERSATION_TAG_COMMAND,
                    self._texts.telegram_create_tag_command_description,
                ),
                BotCommand(self._REPORT_COMMAND, self._texts.telegram_report_command_description),
            ]
        )

    async def start(self) -> None:
        assert self._telegram_app.updater
        await self._telegram_app.updater.start_polling()
        await self._telegram_app.start()

    async def _handle_new_conversation_event(self, event: ConversationEvent) -> None:
        await self._send_or_edit_new_conversation_notifications(event.conversation)

    async def _handle_new_unassigned_message_from_customer_event(
        self, event: NewUnassignedMessageFromCustomerEvent
    ) -> None:
        if len(event.conversation.messages) == 1:
            # Already handled in `_handle_new_conversation_event()`.
            return
        await self._send_or_edit_new_conversation_notifications(event.conversation)

    async def _handle_conversation_assignment_event(self, event: ConversationEvent) -> None:
        await self._send_or_edit_new_conversation_notifications(event.conversation)

    async def _handle_conversation_resolution_event(self, event: ConversationEvent) -> None:
        await self._send_or_edit_new_conversation_notifications(event.conversation)

    async def _handle_conversation_tags_event(self, event: ConversationTagEvent) -> None:
        await self._send_or_edit_new_conversation_notifications(event.conversation)

    async def _handle_tag_event(self, _: TagEvent) -> None:
        # Refresh keyboards for all NEW conversations.
        messages = await self._storage.get_messages(
            kind=TelegramMessageKind.NEW_CONVERSATION_NOTIFICATION
        )
        conversation_ids = list(
            {message.conversation_id for message in messages if message.conversation_id is not None}
        )
        conversations = await self._backend.get_conversations(conversation_ids)
        conv_by_id = {conv.id: conv for conv in conversations}
        all_tags = await self._backend.get_all_tags()
        await flat_gather(
            self._update_new_conversation_notification(message, conv, all_tags, keyboard_only=True)
            for message in messages
            for conv in (conv_by_id[message.conversation_id],)
            if conv.state == ConversationState.NEW
        )

    async def _send_new_conversation_notification(
        self,
        group: TelegramChat,
        conversation: Conversation,
        all_tags: List[Tag],
    ) -> None:
        message = await self._send_placeholder_message(
            group,
            self._texts.telegram_new_conversation_notification_placeholder,
            TelegramMessageKind.NEW_CONVERSATION_NOTIFICATION,
            conversation_id=conversation.id,
        )
        await self._update_new_conversation_notification(message, conversation, all_tags)

    async def _send_placeholder_message(
        self,
        group: TelegramChat,
        text: str,
        kind: TelegramMessageKind,
        conversation_id: Optional[Any] = None,
    ) -> TelegramMessage:
        message = await self._telegram_bot.send_message(group.telegram_chat_id, text)
        return await self._storage.insert_message(
            self._telegram_bot.id,
            group,
            message.message_id,
            kind,
            conversation_id=conversation_id,
        )

    async def _send_or_edit_new_conversation_notifications(
        self, conversation: Conversation
    ) -> None:
        # For this conversation, there are some notifications in some chats.
        #
        # In those chats where newer messages with non-NEW conversation
        # notifications are present, we should delete the old notification and
        # create a new one to avoid losing the notification in the chat history.
        #
        # In other chats (where the notification is still on top or at least
        # among other NEW notifications), we should update the existing notification.
        messages = await self._storage.get_messages(
            TelegramMessageKind.NEW_CONVERSATION_NOTIFICATION,
            conversation_id=conversation.id,
        )
        chats = await self._storage.get_chats_by_role(
            TelegramChatRole.NEW_CONVERSATION_NOTIFICATIONS
        )
        group_ids = {group.telegram_chat_id for group in chats}
        newer_messages = await self._storage.get_newer_messages_of_kind(messages)
        newer_message_conversation_ids = {message.conversation_id for message in newer_messages}
        conversations = await self._backend.get_conversations(list(newer_message_conversation_ids))
        not_new_conversation_ids = {
            conv.id for conv in conversations if conv.state != ConversationState.NEW
        }
        now_new_group_ids = [
            message.chat.telegram_chat_id
            for message in newer_messages
            if message.conversation_id in not_new_conversation_ids
        ]

        messages_to_delete: List[TelegramMessage] = []
        messages_to_update: List[TelegramMessage] = []
        for message in messages:
            if message.chat.telegram_chat_id in now_new_group_ids:
                messages_to_delete.append(message)
            else:
                messages_to_update.append(message)
                group_ids.remove(message.chat.telegram_chat_id)

        chats_to_send_to = [group for group in chats if group.telegram_chat_id in group_ids]

        all_tags = await self._backend.get_all_tags()

        await asyncio.gather(
            flat_gather(
                self._telegram_bot.delete_message(
                    chat_id=message.chat.telegram_chat_id,
                    message_id=message.telegram_message_id,
                )
                for message in messages_to_delete
            ),
            flat_gather(
                self._send_new_conversation_notification(group, conversation, all_tags)
                for group in chats_to_send_to
            ),
            flat_gather(
                self._update_new_conversation_notification(message, conversation, all_tags)
                for message in messages_to_update
            ),
        )

    async def _update_new_conversation_notification(
        self,
        message: TelegramMessage,
        conversation: Conversation,
        all_tags: List[Tag],
        keyboard_only: bool = False,
    ) -> None:
        present_tag_ids = {tag.id for tag in conversation.tags}

        assign_buttons = (
            [[self._make_assign_to_me_button(conversation)]]
            if conversation.state == ConversationState.NEW
            else []
        )
        tag_buttons = [
            self._make_remove_tag_button(conversation, tag)
            if tag.id in present_tag_ids
            else self._make_add_tag_button(conversation, tag)
            for tag in all_tags
        ]
        inline_buttons = [*assign_buttons, *arrange_buttons(tag_buttons)]
        inline_keyboard = InlineKeyboardMarkup(inline_buttons) if inline_buttons else None

        if keyboard_only:
            try:
                await self._telegram_bot.edit_message_reply_markup(
                    message.chat.telegram_chat_id,
                    message.telegram_message_id,
                    reply_markup=inline_keyboard,
                )
                return
            except BadRequest as exc:
                if "Message is not modified" not in str(exc):
                    raise

        text = self._texts.compose_telegram_conversation_notification(conversation)
        try:
            await self._telegram_bot.edit_message_text(
                text.text,
                message.chat.telegram_chat_id,
                message.telegram_message_id,
                parse_mode=text.parse_mode,
                reply_markup=inline_keyboard,
            )
        except BadRequest as exc:
            if "Message is not modified" not in str(exc):
                raise

    def _make_assign_to_me_button(self, conversation: Conversation) -> InlineKeyboardButton:
        return InlineKeyboardButton(
            self._texts.telegram_assign_to_me_button_text,
            callback_data=encode_callback_data(
                {"a": CallbackActionKind.ASSIGN_TO_ME, "c": conversation.id}
            ),
        )

    def _make_add_tag_button(self, conversation: Conversation, tag: Tag) -> InlineKeyboardButton:
        return InlineKeyboardButton(
            text=self._texts.compose_add_tag_button_text(tag),
            callback_data=encode_callback_data(
                {"a": CallbackActionKind.ADD_CONVERSATION_TAG, "c": conversation.id, "t": tag.id}
            ),
        )

    def _make_remove_tag_button(self, conversation: Conversation, tag: Tag) -> InlineKeyboardButton:
        return InlineKeyboardButton(
            text=self._texts.compose_remove_tag_button_text(tag),
            callback_data=encode_callback_data(
                {"a": CallbackActionKind.REMOVE_CONVERSATION_TAG, "c": conversation.id, "t": tag.id}
            ),
        )

    async def _handle_start_command(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert (
            update.effective_chat
        ), "command update with `ChatType.PRIVATE` filter should have `effective_chat`"
        assert (
            update.effective_user
        ), "command update with `ChatType.PRIVATE` filter should have `effective_user`"
        try:
            await self._backend.identify_agent(make_agent_identification(update.effective_user))
        except AgentNotFound:
            answer = self._texts.telegram_manager_permission_denied_message
        else:
            answer = self._texts.telegram_manager_start_message
        await context.bot.send_message(update.effective_chat.id, answer)

    @send_text_answer
    async def _handle_create_conversation_tag_command(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> str:
        assert update.message, "command update with `TEXT` filter should have `message`"
        assert update.message.text, "command update with `TEXT` filter should have `message.text`"
        assert update.effective_user, "command update should have `effective_user`"
        try:
            agent = await self._backend.identify_agent(
                make_agent_identification(update.effective_user)
            )
        except AgentNotFound:
            return self._texts.telegram_manager_permission_denied_message

        tag_name = update.message.text.removeprefix(
            f"/{self._CREATE_CONVERSATION_TAG_COMMAND}"
        ).strip()
        if not tag_name:
            return self._texts.telegram_create_tag_usage_message

        try:
            await self._backend.create_tag(tag_name, agent)
            return self._texts.telegram_tag_successfully_created_message
        except TagAlreadyExists:
            return self._texts.telegram_tag_already_exists_message
        except PermissionDenied:
            return self._texts.telegram_create_tag_permission_denied_message

    async def _handle_report_command(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert update.message, "command update with `TEXT` filter should have `message`"
        assert update.message.text, "command update with `TEXT` filter should have `message.text`"
        assert update.effective_chat, "command update should have `effective_chat`"
        assert update.effective_user, "command update should have `effective_user`"

        chat_id = update.effective_chat.id

        try:
            await self._backend.identify_agent(make_agent_identification(update.effective_user))
        except AgentNotFound:
            await context.bot.send_message(
                chat_id=chat_id, text=self._texts.telegram_manager_permission_denied_message
            )

        last_progress = "0%"
        last_progress_update = perf_counter()
        message = await context.bot.send_message(
            chat_id=chat_id,
            text=self._texts.telegram_report_message_placeholder.format(progress=last_progress),
        )

        async def progress_callback(progress: float) -> None:
            nonlocal last_progress
            nonlocal last_progress_update
            now = perf_counter()
            curr_progress = f"{100.0 * progress:.1f}%"
            if now - last_progress_update > 1.0 and curr_progress != last_progress:
                await context.bot.edit_message_text(
                    chat_id=chat_id,
                    message_id=message.message_id,
                    text=self._texts.telegram_report_message_placeholder.format(
                        progress=curr_progress
                    ),
                )
                last_progress = curr_progress
                last_progress_update = now

        report = await self._reporter.report(progress_callback)
        await context.bot.delete_message(chat_id=chat_id, message_id=message.message_id)
        await context.bot.send_message(
            chat_id=chat_id, text=self._texts.telegram_report_message.format(report=report)
        )
        # TODO send xlsx report

    async def _handle_callback_query(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert update.callback_query, "callback query update should have `callback_query`"
        assert update.effective_chat, "callback query update should have `effective_chat`"
        assert update.effective_user, "callback query update should have `effective_user`"
        if not update.callback_query.data:
            # No idea how to handle this update.
            return
        await self._create_or_update_agent(update.effective_chat, update.effective_user)
        callback_data = decode_callback_data(update.callback_query.data)
        action = callback_data["a"]
        if action == CallbackActionKind.ASSIGN_TO_ME:
            conversation_id = callback_data["c"]
            agent = await self._backend.identify_agent(
                make_agent_identification(update.effective_user)
            )
            await self._backend.assign_agent(agent, agent, conversation_id)
        elif action in (
            CallbackActionKind.ADD_CONVERSATION_TAG,
            CallbackActionKind.REMOVE_CONVERSATION_TAG,
        ):
            conversation_id = callback_data["c"]
            conversation = await self._backend.get_conversation(conversation_id)
            all_tags = await self._backend.get_all_tags()
            tag_id = callback_data["t"]
            tag = next(tag for tag in all_tags if tag.id == tag_id)
            if action == CallbackActionKind.ADD_CONVERSATION_TAG:
                await self._backend.add_tag_to_conversation(conversation, tag)
            else:
                await self._backend.remove_tag_from_conversation(conversation, tag)
        else:
            logger.info(f"Manager frontend received unsupported callback action {action!r}")

    async def _create_or_update_agent(self, effective_chat: Chat, effective_user: User) -> None:
        identification = make_agent_identification(effective_user)
        diff = make_agent_diff(effective_user)
        try:
            await self._backend.update_agent(identification, diff)
        except AgentNotFound:
            group = await self._storage.get_chat(effective_chat.id)
            if TelegramChatRole.AGENTS not in group.roles:
                return
            await self._backend.create_or_update_agent(identification, diff)

    async def _handle_agents_command(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert update.effective_chat, "command update should have `effective_chat`"
        assert update.effective_user, "command update should have `effective_user`"
        await self._backend.identify_agent(make_agent_identification(update.effective_user))
        await self._storage.create_or_update_chat(update.effective_chat.id)
        await self._storage.add_chat_roles(
            update.effective_chat.id,
            TelegramChatRole.AGENTS,
            TelegramChatRole.NEW_CONVERSATION_NOTIFICATIONS,
        )
        await context.bot.send_message(
            update.effective_chat.id, self._texts.telegram_agents_command_success_message
        )

    async def _handle_send_new_conversations_command(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert update.effective_chat, "command update should have `effective_chat`"
        assert update.effective_user, "command update should have `effective_user`"
        await self._backend.identify_agent(make_agent_identification(update.effective_user))
        await self._storage.create_or_update_chat(update.effective_chat.id)
        await self._storage.add_chat_roles(
            update.effective_chat.id, TelegramChatRole.NEW_CONVERSATION_NOTIFICATIONS
        )
        await context.bot.send_message(
            update.effective_chat.id,
            self._texts.telegram_send_new_conversations_command_success_message,
        )

    async def _handle_new_chat_members(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert (
            update.effective_chat
        ), "update with NEW_CHAT_MEMBERS filter should have `effective_chat`"
        assert (
            update.message and update.message.new_chat_members
        ), "update with NEW_CHAT_MEMBERS filter should have `message.new_chat_members`"
        group = await self._storage.create_or_update_chat(update.effective_chat.id)
        if TelegramChatRole.AGENTS not in group.roles:
            return
        await flat_gather(
            self._backend.create_or_update_agent(make_agent_identification(user))
            for user in update.message.new_chat_members
            if not user.is_bot
        )

    async def _handle_left_chat_member(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> None:
        assert (
            update.effective_chat
        ), "update with LEFT_CHAT_MEMBERS filter should have `effective_chat`"
        assert (
            update.message and update.message.left_chat_member
        ), "update with LEFT_CHAT_MEMBERS filter should have `message.left_chat_member`"

        user = update.message.left_chat_member
        if user.is_bot:
            return

        group = await self._storage.create_or_update_chat(update.effective_chat.id)
        if TelegramChatRole.AGENTS not in group.roles:
            return

        try:
            agent = await self._backend.identify_agent(make_agent_identification(user))
        except AgentNotFound:
            return

        if await self._helper.check_belongs_to_agent_chats(user.id):
            return

        await self._backend.deactivate_agent(agent)



import asyncio
import logging
from itertools import groupby
from typing import List, Iterable, Optional, Callable, Awaitable, Mapping

from telegram import (
    Update,
    BotCommand,
    Bot,
    Chat,
)
from telegram.error import BadRequest, Forbidden
from telegram.ext import (
    MessageHandler,
    ContextTypes,
    CommandHandler,
    Application,
    CallbackQueryHandler,
    BaseHandler,
)
from telegram.ext.filters import TEXT, ChatType

from suppgram.backend import Backend
from suppgram.entities import (
    WorkplaceIdentification,
    MessageKind,
    Message,
    NewMessageForAgentEvent,
    Workplace,
    ConversationEvent,
    Customer,
    FINAL_STATES,
    Agent,
    Conversation,
)
from suppgram.errors import ConversationNotFound, AgentNotFound
from suppgram.frontend import (
    AgentFrontend,
)
from suppgram.frontends.telegram.app_manager import TelegramAppManager
from suppgram.frontends.telegram.callback_actions import CallbackActionKind
from suppgram.frontends.telegram.helper import TelegramHelper
from suppgram.frontends.telegram.helpers import (
    is_chat_not_found,
    is_blocked_by_user,
    make_pagination_keyboard,
    paginate_texts,
    decode_callback_data,
)
from suppgram.frontends.telegram.identification import (
    make_agent_identification,
    make_workplace_identification,
    make_agent_diff,
)
from suppgram.frontends.telegram.storage import (
    TelegramStorage,
    TelegramChat,
    TelegramMessageKind,
    TelegramMessage,
)
from suppgram.helpers import flat_gather
from suppgram.texts.interface import TextProvider, Format

logger = logging.getLogger(__name__)


class TelegramAgentFrontend(AgentFrontend):
    _POSTPONE_COMMAND = "postpone"
    _RESOLVE_COMMAND = "resolve"

    def __init__(
        self,
        agent_bot_tokens: List[str],
        manager_bot_token: Optional[str],
        app_manager: TelegramAppManager,
        backend: Backend,
        helper: TelegramHelper,
        storage: TelegramStorage,
        texts: TextProvider,
    ):
        self._backend = backend
        self._helper = helper
        self._storage = storage
        self._texts = texts
        self._telegram_apps: List[Application] = []
        for token in agent_bot_tokens:
            app = app_manager.get_app(token)
            app.add_handlers(self._make_handlers())
            self._telegram_apps.append(app)
        self._telegram_app_by_bot_id: Mapping[int, Application] = {}
        self._manager_bot: Optional[Bot] = (
            app_manager.get_app(manager_bot_token).bot if manager_bot_token else None
        )
        self._backend.on_conversation_assignment.add_handler(self._handle_conversation_assignment)
        self._backend.on_new_message_for_agent.add_batch_handler(
            self._handle_new_message_for_agent_events
        )

    def _make_handlers(self) -> List[BaseHandler]:
        return [
            CommandHandler("start", self._handle_start_command, filters=ChatType.PRIVATE),
            CommandHandler(
                self._POSTPONE_COMMAND,
                self._handle_postpone_command,
                filters=ChatType.PRIVATE,
            ),
            CommandHandler(
                self._RESOLVE_COMMAND,
                self._handle_resolve_command,
                filters=ChatType.PRIVATE,
            ),
            MessageHandler(ChatType.PRIVATE & TEXT, self._handle_text_message),
            MessageHandler(ChatType.PRIVATE & ~TEXT, self._handle_unsupported_message),
            CallbackQueryHandler(self._handle_callback_query),
        ]

    async def initialize(self):
        await super().initialize()
        await flat_gather(app.initialize() for app in self._telegram_apps)
        await flat_gather(self._set_commands(app) for app in self._telegram_apps)

    async def _set_commands(self, app: Application):
        postpone = BotCommand(
            self._POSTPONE_COMMAND,
            self._texts.telegram_postpone_command_description,
        )
        resolve = BotCommand(
            self._RESOLVE_COMMAND,
            self._texts.telegram_resolve_command_description,
        )
        await app.bot.set_my_commands([postpone, resolve])

    async def start(self):
        await flat_gather(app.updater.start_polling() for app in self._telegram_apps)
        await flat_gather(app.start() for app in self._telegram_apps)

    async def _handle_start_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.effective_chat, "command update should have `effective_chat`"
        assert update.effective_user, "command update should have `effective_user`"
        workplace_identification = make_workplace_identification(update, update.effective_user)
        try:
            workplace = await self._backend.identify_workplace(workplace_identification)
            agent = workplace.agent
        except AgentNotFound:
            should_create = await self._helper.check_belongs_to_agent_chats(
                update.effective_user.id
            )
            if not should_create:
                await context.bot.send_message(
                    chat_id=update.effective_chat.id,
                    text=self._texts.telegram_manager_permission_denied_message,
                )
                return
            agent_identification = workplace_identification.to_agent_identification()
            agent = await self._backend.create_or_update_agent(
                agent_identification, make_agent_diff(update.effective_user)
            )

        await asyncio.gather(
            self._send_start_message_and_conversation_messages(
                update.effective_chat, workplace_identification, context
            ),
            self._delete_nudges(agent, update),
        )

    async def _send_start_message_and_conversation_messages(
        self,
        effective_chat: Chat,
        workplace_identification: WorkplaceIdentification,
        context: ContextTypes.DEFAULT_TYPE,
    ):
        await context.bot.send_message(
            chat_id=effective_chat.id, text=self._texts.telegram_agent_start_message
        )

        try:
            conversation = await self._backend.identify_agent_conversation(workplace_identification)
        except ConversationNotFound:
            pass
        else:
            await self._handle_conversation_assignment(ConversationEvent(conversation))

    async def _delete_nudges(self, agent: Agent, update: Update):
        app = self._get_app_by_bot_id(update.get_bot().id)
        messages = await self._storage.get_messages(
            TelegramMessageKind.NUDGE_TO_START_BOT_NOTIFICATION,
            agent_id=agent.id,
            telegram_bot_username=app.bot.username,
        )
        await flat_gather(self._helper.delete_message_if_exists(message) for message in messages)
        await self._storage.delete_messages(messages)

    async def _handle_postpone_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        await self._handle_conversation_command(
            update, context, self._backend.postpone_conversation
        )

    async def _handle_resolve_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        await self._handle_conversation_command(update, context, self._backend.resolve_conversation)

    async def _handle_conversation_command(
        self,
        update: Update,
        context: ContextTypes.DEFAULT_TYPE,
        command: Callable[[Agent, Conversation], Awaitable[None]],
    ):
        assert update.effective_chat, "command update should have `effective_chat`"
        assert update.effective_user, "command update should have `effective_user`"
        try:
            agent = await self._backend.identify_agent(
                make_agent_identification(update.effective_user)
            )
        except AgentNotFound:
            answer = self._texts.telegram_manager_permission_denied_message
            await context.bot.send_message(update.effective_chat.id, answer)
            return

        try:
            conversation = await self._backend.identify_agent_conversation(
                make_workplace_identification(update, update.effective_user)
            )
        except ConversationNotFound:
            answer = self._texts.telegram_workplace_is_not_assigned_message
            await context.bot.send_message(update.effective_chat.id, answer)
            return

        await command(agent, conversation)

    async def _handle_text_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.effective_chat, "message update should have `effective_chat`"
        assert update.effective_user, "message update should have `effective_user`"
        assert update.message, "message update should have `message`"
        try:
            conversation = await self._backend.identify_agent_conversation(
                WorkplaceIdentification(
                    telegram_user_id=update.effective_user.id,
                    telegram_bot_id=context.bot.bot.id,
                )
            )
        except ConversationNotFound:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text=self._texts.telegram_workplace_is_not_assigned_message,
            )
            return
        await self._backend.process_message(
            conversation,
            Message(
                kind=MessageKind.FROM_AGENT,
                time_utc=update.message.date,
                text=update.message.text,
            ),
        )

    async def _handle_unsupported_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.effective_chat, "message update should have `effective_chat`"
        assert update.effective_user, "message update should have `effective_user`"
        assert update.message, "message update should have `message`"
        try:
            await self._backend.identify_agent_conversation(
                WorkplaceIdentification(
                    telegram_user_id=update.effective_user.id,
                    telegram_bot_id=context.bot.bot.id,
                )
            )
        except ConversationNotFound:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text=self._texts.telegram_workplace_is_not_assigned_message,
            )
            return
        await context.bot.send_message(
            chat_id=update.effective_chat.id,
            text=self._texts.telegram_agent_unsupported_message_content,
            reply_to_message_id=update.message.message_id,
        )

    async def _handle_callback_query(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.callback_query, "callback query update should have `callback_query`"
        assert update.effective_chat, "callback query update should have `effective_chat`"
        assert update.effective_user, "callback query update should have `effective_user`"
        if not update.callback_query.data:
            # No idea how to handle this update.
            return

        agent = await self._backend.identify_agent(make_agent_identification(update.effective_user))
        callback_data = decode_callback_data(update.callback_query.data)
        action = callback_data.get("a")
        if action is None:
            return
        if action == CallbackActionKind.PAGE:
            chat = await self._storage.get_chat(callback_data["c"])
            message = await self._storage.get_message(chat, callback_data["m"])
            if message.agent_id != agent.id:
                # Sorry, this message is not for you!
                # Probably someone's playing with the callback query payloads.
                return
            customer = await self._backend.create_or_update_customer(
                message.customer_identification,
            )
            await self._update_previous_conversations_message(customer, message, callback_data["p"])
        else:
            logger.info(f"Agent frontend received unsupported callback action {action!r}")

    async def _handle_conversation_assignment(self, event: ConversationEvent):
        conversation = event.conversation
        workplace = conversation.assigned_workplace
        assert (
            workplace
        ), "conversation should have assigned workplace upon conversation assignment event"
        try:
            await self._send_customer_profile(workplace, conversation.customer)
            await self._send_previous_conversations_message(workplace, conversation.customer)
            await self._send_new_messages(workplace, conversation.messages)
        except (BadRequest, Forbidden) as exc:
            if is_chat_not_found(exc) or is_blocked_by_user(exc):
                await self._nudge_to_start_bot(workplace)
                return
            raise

    async def _handle_new_message_for_agent_events(self, events: List[NewMessageForAgentEvent]):
        for _, batch_iter in groupby(
            events, lambda event: (event.workplace.agent.id, event.workplace.id)
        ):
            batch = list(batch_iter)
            workplace = batch[0].workplace
            messages = [event.message for event in batch]
            await self._send_new_messages(workplace, messages)

    async def _send_customer_profile(self, workplace: Workplace, customer: Customer):
        if workplace.telegram_user_id is None or workplace.telegram_bot_id is None:
            return
        app = self._get_app_by_bot_id(workplace.telegram_bot_id)
        profile = self._texts.compose_customer_profile(
            customer, Format.get_formats_supported_by_telegram()
        )
        await app.bot.send_message(
            chat_id=workplace.telegram_user_id, text=profile.text, parse_mode=profile.parse_mode
        )

    async def _send_previous_conversations_message(self, workplace: Workplace, customer: Customer):
        if workplace.telegram_bot_id is None or workplace.telegram_user_id is None:
            return
        pages = await self._compose_previous_conversations_message_texts(customer)
        if not pages:
            return
        app = self._get_app_by_bot_id(workplace.telegram_bot_id)
        message = await app.bot.send_message(chat_id=workplace.telegram_user_id, text=pages[0])
        chat = await self._storage.create_or_update_chat(workplace.telegram_user_id)
        tmessage = await self._storage.insert_message(
            workplace.telegram_bot_id,
            chat,
            message.message_id,
            TelegramMessageKind.CUSTOMER_MESSAGE_HISTORY,
            agent_id=workplace.agent.id,
            customer_id=customer.id,
        )
        if (keyboard := make_pagination_keyboard(tmessage, len(pages), 0)) is not None:
            await app.bot.edit_message_reply_markup(
                chat_id=message.chat_id, message_id=message.message_id, reply_markup=keyboard
            )

    async def _update_previous_conversations_message(
        self, customer: Customer, message: TelegramMessage, page_idx: int
    ):
        app = self._get_app_by_bot_id(message.telegram_bot_id)
        pages = await self._compose_previous_conversations_message_texts(customer)
        if not (0 <= page_idx < len(pages)):
            return
        keyboard = make_pagination_keyboard(message, len(pages), page_idx)
        await app.bot.edit_message_text(
            pages[page_idx],
            message.chat.telegram_chat_id,
            message.telegram_message_id,
            reply_markup=keyboard,
        )

    async def _compose_previous_conversations_message_texts(self, customer: Customer) -> List[str]:
        conversations = await self._backend.get_customer_conversations(customer)
        previous_conversations = [conv for conv in conversations if conv.state in FINAL_STATES]
        messages = sorted(
            [message for conv in previous_conversations for message in conv.messages],
            key=lambda m: m.time_utc,
        )
        if not messages:
            return []
        formatted_messages = self._format_previous_messages(messages)
        pages = self._paginate_formatted_messages(formatted_messages)
        return list(pages)

    def _format_previous_messages(self, messages: Iterable[Message]) -> Iterable[str]:
        for message in messages:
            yield self._texts.format_history_message(message)

    def _paginate_formatted_messages(self, messages: Iterable[str]) -> Iterable[str]:
        yield from paginate_texts(
            prefix=self._texts.message_history_title,
            texts=messages,
            suffix="",
            delimiter="\n",
            max_page_lines=15,
            max_page_chars=1000,
        )

    async def _send_new_messages(self, workplace: Workplace, messages: List[Message]):
        if workplace.telegram_user_id is None or workplace.telegram_bot_id is None:
            return

        app = self._get_app_by_bot_id(workplace.telegram_bot_id)
        texts = self._group_messages(messages)
        try:
            for text in texts:
                await app.bot.send_message(chat_id=workplace.telegram_user_id, text=text)
        except (BadRequest, Forbidden) as exc:
            if is_chat_not_found(exc) or is_blocked_by_user(exc):
                await self._nudge_to_start_bot(workplace)
                return
            raise

    def _get_app_by_bot_id(self, telegram_bot_id: int) -> Application:
        if not self._telegram_app_by_bot_id:
            self._telegram_app_by_bot_id = {app.bot.id: app for app in self._telegram_apps}
        return self._telegram_app_by_bot_id[telegram_bot_id]

    async def _nudge_to_start_bot(self, workplace: Workplace):
        assert workplace.telegram_user_id is not None, "should be called for Telegram workplaces"

        agent_chats = await self._helper.check_belongs_to_agent_chats(workplace.telegram_user_id)
        await flat_gather(self._nudge_to_start_bot_in_chat(workplace, chat) for chat in agent_chats)

    async def _nudge_to_start_bot_in_chat(self, workplace: Workplace, chat: TelegramChat):
        assert workplace.telegram_bot_id is not None, "should be called for Telegram workplaces"

        if self._manager_bot is None:
            return
        app = self._get_app_by_bot_id(workplace.telegram_bot_id)
        bot_username = app.bot.username
        text = self._texts.compose_nudge_to_start_bot_notification(workplace.agent, bot_username)
        message = await self._manager_bot.send_message(
            chat_id=chat.telegram_chat_id, text=text.text, parse_mode=text.parse_mode
        )
        await self._storage.insert_message(
            workplace.telegram_bot_id,
            chat,
            message.message_id,
            TelegramMessageKind.NUDGE_TO_START_BOT_NOTIFICATION,
            agent_id=workplace.agent.id,
            telegram_bot_username=bot_username,
        )

    def _group_messages(self, messages: Iterable[Message]) -> Iterable[str]:
        yield from paginate_texts(
            prefix="",
            texts=(self._texts.format_history_message(message) for message in messages),
            suffix="",
            # When unspecified, `max_page_lines` and `max_page_chars` are assumed to be
            # roughly equal to Telegram's own limits on message length. We are fine with
            # that here, because we actually want to send as few messages as possible
            # to avoid hitting Telegram RPS limit.
        )



from logging import StreamHandler, LogRecord
from typing import Mapping


class ConfidentialStreamHandler(StreamHandler):
    def __init__(self, stream=None, replacements: Mapping[str, str] = {}):
        super().__init__(stream)
        self._replacements = replacements

    def format(self, record: LogRecord) -> str:
        result = super().format(record)
        for k, v in self._replacements.items():
            result = result.replace(k, v)
        return result



from suppgram.entities import WorkplaceIdentification, AgentIdentification, CustomerIdentification


class NoStorageSpecified(Exception):
    pass


class NoFrontendSpecified(Exception):
    pass


class AgentException(Exception):
    def __init__(self, identification: AgentIdentification):
        self.identification = identification


class AgentEmptyIdentification(AgentException):
    def __str__(self):
        return (
            f"received agent identification {self.identification} without supported non-null fields"
        )


class AgentNotFound(AgentException):
    pass


class AgentDeactivated(AgentException):
    pass


class WorkplaceException(Exception):
    def __init__(self, identification: WorkplaceIdentification):
        self.identification = identification


class WorkplaceEmptyIdentification(WorkplaceException):
    def __str__(self):
        return f"received workplace identification {self.identification} without supported non-null fields"


class WorkplaceNotFound(WorkplaceException):
    pass


class TagAlreadyExists(Exception):
    pass


class ConversationAlreadyAssigned(Exception):
    pass


class CustomerException(Exception):
    def __init__(self, identification: CustomerIdentification):
        self.identification = identification


class CustomerNotFound(CustomerException):
    pass


class CustomerEmptyIdentification(CustomerException):
    def __str__(self):
        return f"received customer identification {self.identification} without supported non-null fields"


class ConversationNotFound(Exception):
    pass


class PermissionDenied(Exception):
    pass


class DataNotFetched(RuntimeError):
    pass



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Optional, List, Dict, Set, Awaitable, Callable

from suppgram.entities import Event, MessageKind, EventKind
from suppgram.helpers import aenumerate
from suppgram.storage import Storage


@dataclass(frozen=True)
class AgentAnalytics:
    agent_id: Any
    telegram_user_id: Optional[int]

    total_assigned: int
    total_resolved: int
    average_customer_rating: float
    average_assignment_to_resolution_time_min: float


@dataclass(frozen=True)
class ConversationAnalytics:
    conversation_id: Any
    customer_id: Any
    last_assigned_agent_id: Optional[Any]
    customer_rating: Optional[int]

    creation_time_utc: datetime
    first_assignment_time_utc: Optional[datetime]
    resolve_time_utc: Optional[datetime]

    last_message_kind: MessageKind
    last_message_time_utc: datetime


@dataclass(frozen=True)
class Report:
    agents: List[AgentAnalytics]
    conversations: List[ConversationAnalytics]
    events: List[Event]

    average_start_to_first_response_time_min: float
    average_start_to_resolution_time_min: float
    average_customer_rating: float


ProgressCallback = Callable[[float], Awaitable[None]]


@dataclass
class _ConversationState:
    conversation_id: Any
    last_assigned_agent_id: Optional[Any] = None

    creation_time_utc: Optional[datetime] = None
    first_assignment_time_utc: Optional[datetime] = None
    first_response_time_utc: Optional[datetime] = None
    last_assignment_time_utc: Optional[datetime] = None
    resolve_time_utc: Optional[datetime] = None

    last_message_kind: Optional[MessageKind] = None
    last_message_time_utc: Optional[datetime] = None


class Reporter:
    def __init__(self, storage: Storage):
        self._storage = storage

    async def report(self, progress_callback: Optional[ProgressCallback]) -> Report:
        if progress_callback is not None:
            await progress_callback(0.0)
        total_events = await self._storage.count_all_events()
        total_conversations = await self._storage.count_all_conversations()
        total_steps = max(1, total_events + total_conversations)
        # Might also count agents, but there are probably going
        # to be very few compared to conversations and events.

        conversation_states: Dict[Any, _ConversationState] = {}
        agent_assignments: Dict[Any, Set[Any]] = defaultdict(set)
        agent_resolutions: Dict[Any, Set[Any]] = defaultdict(set)
        agent_ratings: Dict[Any, List[int]] = defaultdict(list)
        agent_resolution_times: Dict[Any, List[timedelta]] = defaultdict(list)

        def _conv_state(conv_id: Any) -> _ConversationState:
            conversation_states.setdefault(conv_id, _ConversationState(conversation_id=conv_id))
            return conversation_states[conv_id]

        events: List[Event] = []
        async for i, event in aenumerate(self._storage.find_all_events()):
            events.append(event)
            conv_state = _conv_state(event.conversation_id)

            if event.kind == EventKind.AGENT_ASSIGNED:
                conv_state.last_assigned_agent_id = event.agent_id
                if not conv_state.first_assignment_time_utc:
                    conv_state.first_assignment_time_utc = event.time_utc
                conv_state.last_assignment_time_utc = event.time_utc
                agent_assignments[event.agent_id].add(event.conversation_id)

            elif event.kind == EventKind.CONVERSATION_POSTPONED:
                pass

            elif event.kind == EventKind.CONVERSATION_RATED:
                # We'll only take into account final ratings stored in conversations.
                # Rating will be attributed to the last agent assigned to a conversation.
                pass

            elif event.kind == EventKind.CONVERSATION_RESOLVED:
                conv_state.resolve_time_utc = event.time_utc
                agent_resolutions[conv_state.last_assigned_agent_id].add(event.conversation_id)
                if conv_state.last_assignment_time_utc:
                    agent_resolution_times[conv_state.last_assigned_agent_id].append(
                        event.time_utc - conv_state.last_assignment_time_utc
                    )

            elif event.kind == EventKind.CONVERSATION_STARTED:
                conv_state.creation_time_utc = event.time_utc

            elif event.kind == EventKind.CONVERSATION_TAG_ADDED:
                pass

            elif event.kind == EventKind.CONVERSATION_TAG_REMOVED:
                pass

            elif event.kind == EventKind.MESSAGE_SENT:
                if (
                    event.message_kind == MessageKind.FROM_AGENT
                    and not conv_state.first_response_time_utc
                ):
                    conv_state.first_response_time_utc = event.time_utc
                if event.message_kind in (MessageKind.FROM_AGENT, MessageKind.FROM_CUSTOMER):
                    conv_state.last_message_kind = event.message_kind
                    conv_state.last_message_time_utc = event.time_utc

            if progress_callback is not None:
                await progress_callback((i + 1) / total_steps)

        conversations: List[ConversationAnalytics] = []
        async for i, conv in aenumerate(self._storage.find_all_conversations()):
            conv_state = _conv_state(conv.id)
            if (
                not conv_state.creation_time_utc
                or not conv_state.last_message_kind
                or not conv_state.last_message_time_utc
            ):
                # Not a single message within conversation — ignoring it.
                continue

            if conv.customer_rating is not None and conv_state.last_assigned_agent_id is not None:
                agent_ratings[conv_state.last_assigned_agent_id].append(conv.customer_rating)
            conversations.append(
                ConversationAnalytics(
                    conversation_id=conv.id,
                    customer_id=conv.customer.id,
                    last_assigned_agent_id=conv_state.last_assigned_agent_id,
                    customer_rating=conv.customer_rating,
                    creation_time_utc=conv_state.creation_time_utc,
                    first_assignment_time_utc=conv_state.first_assignment_time_utc,
                    resolve_time_utc=conv_state.resolve_time_utc,
                    last_message_kind=conv_state.last_message_kind,
                    last_message_time_utc=conv_state.last_message_time_utc,
                )
            )

            if progress_callback is not None:
                await progress_callback((total_events + i + 1) / total_steps)

        agents = [
            AgentAnalytics(
                agent_id=agent.id,
                telegram_user_id=agent.telegram_user_id,
                total_assigned=len(agent_assignments[agent.id]),
                total_resolved=len(agent_resolutions[agent.id]),
                average_customer_rating=sum(agent_ratings[agent.id]) / len(agent_ratings[agent.id])
                if agent_ratings[agent.id]
                else float("nan"),
                average_assignment_to_resolution_time_min=sum(
                    td.total_seconds() / 60.0 for td in agent_resolution_times[agent.id]
                )
                / len(agent_resolution_times[agent.id])
                if agent_resolution_times[agent.id]
                else float("nan"),
            )
            async for agent in self._storage.find_all_agents()
        ]

        first_response_times_min = [
            (conv_state.first_response_time_utc - conv_state.creation_time_utc).total_seconds()
            / 60.0
            for conv_state in conversation_states.values()
            if conv_state.creation_time_utc and conv_state.first_response_time_utc
        ]
        resolution_times_min = [
            (conv_state.resolve_time_utc - conv_state.creation_time_utc).total_seconds() / 60.0
            for conv_state in conversation_states.values()
            if conv_state.creation_time_utc and conv_state.resolve_time_utc
        ]
        ratings = [
            conv.customer_rating for conv in conversations if conv.customer_rating is not None
        ]

        return Report(
            agents=agents,
            conversations=conversations,
            events=events,
            average_start_to_first_response_time_min=sum(first_response_times_min)
            / len(first_response_times_min)
            if first_response_times_min
            else float("nan"),
            average_start_to_resolution_time_min=sum(resolution_times_min)
            / len(resolution_times_min)
            if resolution_times_min
            else float("nan"),
            average_customer_rating=sum(ratings) / len(ratings) if ratings else float("nan"),
        )



import asyncio
import uuid
from datetime import datetime, timezone

import aioconsole

from suppgram.backend import Backend
from suppgram.entities import (
    NewMessageForCustomerEvent,
    MessageKind,
    CustomerIdentification,
    Message,
)
from suppgram.frontend import CustomerFrontend
from suppgram.texts.interface import TextProvider


class ShellCustomerFrontend(CustomerFrontend):
    """
    Allows customers to access the support system via terminal.

    Useful for debug purposes.
    """

    def __init__(self, backend: Backend, texts: TextProvider):
        """
        Arguments:
            backend: used backend instance.
            texts: texts provider.
        """
        self._uuid = uuid.uuid4()
        self._backend = backend
        self._texts = texts

        backend.on_new_message_for_customer.add_handler(self._handle_new_message_for_customer_event)

    async def start(self):
        asyncio.create_task(self._run())

    async def _run(self):
        print(self._texts.telegram_customer_start_message)
        while True:
            text = await aioconsole.ainput("You: ")
            conversation = await self._backend.identify_customer_conversation(
                CustomerIdentification(shell_uuid=self._uuid)
            )
            await self._backend.process_message(
                conversation,
                Message(
                    kind=MessageKind.FROM_CUSTOMER,
                    time_utc=datetime.now(timezone.utc),
                    text=text,
                ),
            )

    async def _handle_new_message_for_customer_event(self, event: NewMessageForCustomerEvent):
        text = event.message.text
        if event.message.kind == MessageKind.RESOLVED:
            text = self._texts.telegram_customer_conversation_resolved_message
        print("\rAgent:", text, "\nYou: ", end="")



import asyncio
import re
from typing import Iterable, Awaitable, Optional, AsyncIterator, Tuple, TypeVar


async def flat_gather(futures: Iterable[Awaitable]):
    """Like asyncio.gather(), but accepts iterable of
    coroutines rather than multiple arguments. Useful
    in list comprehensions."""
    return await asyncio.gather(*futures)


# The following function is taken from python-telegram-bot
# to avoid unnecessary dependencies.
def escape_markdown(text: str, entity_type: Optional[str] = None) -> str:
    """
    Helper function to escape telegram markup symbols.

    Args:
        text (:obj:`str`): The text.
        entity_type (:obj:`str`, optional): For the entity types ``PRE``, ``CODE`` and the link
            part of ``TEXT_LINKS``, only certain characters need to be escaped in ``MarkdownV2``.
            See the official API documentation for details. Only valid in combination with
            ``version=2``, will be ignored else.
    """
    if entity_type in ["pre", "code"]:
        escape_chars = r"\`"
    elif entity_type == "text_link":
        escape_chars = r"\)"
    else:
        escape_chars = r"_*[]()~`>#+-=|{}.!"

    return re.sub(f"([{re.escape(escape_chars)}])", r"\\\1", text)


T = TypeVar("T")


async def aenumerate(asequence: AsyncIterator[T], start=0) -> AsyncIterator[Tuple[int, T]]:
    """Asynchronously enumerate an async iterator from a given start value."""
    n = start
    async for elem in asequence:
        yield n, elem
        n += 1



from datetime import datetime, timezone
from typing import List, Any, Optional, Mapping, Dict
from uuid import UUID

from sqlalchemy import (
    Integer,
    String,
    ForeignKey,
    DateTime,
    Table,
    Column,
    Enum,
    ColumnElement,
    Boolean,
)
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.orm import (
    declarative_base,
    Mapped,
    mapped_column,
    relationship,
    joinedload,
    selectinload,
)
from sqlalchemy.orm.interfaces import LoaderOption

from suppgram.containers import UnavailableList
from suppgram.entities import (
    Agent as AgentInterface,
    AgentDiff,
    AgentIdentification,
    Conversation as ConversationInterface,
    ConversationDiff,
    ConversationState,
    Customer as CustomerInterface,
    CustomerDiff,
    CustomerIdentification,
    Event as EventInterface,
    EventKind,
    FINAL_STATES,
    Message as ConversaionMessageInterface,
    MessageKind,
    MessageMediaKind,
    SetNone,
    Tag as TagInterface,
    Workplace as WorkplaceInterface,
    WorkplaceIdentification,
)
from suppgram.errors import AgentEmptyIdentification, WorkplaceEmptyIdentification

Base = declarative_base()


class Customer(Base):
    __tablename__ = "suppgram_customers"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    telegram_user_id: Mapped[int] = mapped_column(Integer, nullable=True, unique=True)
    telegram_first_name: Mapped[str] = mapped_column(String, nullable=True)
    telegram_last_name: Mapped[str] = mapped_column(String, nullable=True)
    telegram_username: Mapped[str] = mapped_column(String, nullable=True)
    shell_uuid: Mapped[str] = mapped_column(String, nullable=True, unique=True)
    pubnub_user_id: Mapped[str] = mapped_column(
        String, nullable=True
    )  # TODO unique together with channel
    pubnub_channel_id: Mapped[str] = mapped_column(String, nullable=True)
    conversations: Mapped[List["Conversation"]] = relationship(back_populates="customer")


class Agent(Base):
    __tablename__ = "suppgram_agents"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    deactivated: Mapped[bool] = mapped_column(Boolean, default=False)
    telegram_user_id: Mapped[int] = mapped_column(Integer, nullable=True, unique=True)
    telegram_first_name: Mapped[str] = mapped_column(String, nullable=True)
    telegram_last_name: Mapped[str] = mapped_column(String, nullable=True)
    telegram_username: Mapped[str] = mapped_column(String, nullable=True)
    workplaces: Mapped[List["Workplace"]] = relationship(back_populates="agent")
    created_tags: Mapped[List["Tag"]] = relationship(
        back_populates="created_by"
    )  # `created_tags` is not really needed, but without it mypy terminates with an exception...


class Workplace(Base):
    __tablename__ = "suppgram_workplaces"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    agent_id: Mapped[int] = mapped_column(ForeignKey(Agent.id))
    agent: Mapped[Agent] = relationship(back_populates="workplaces")
    telegram_bot_id: Mapped[int] = mapped_column(Integer)
    conversations: Mapped[List["Conversation"]] = relationship(back_populates="assigned_workplace")


class Tag(Base):
    __tablename__ = "suppgram_tags"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String, nullable=False, unique=True)
    created_at_utc: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)
    created_by_id: Mapped[int] = mapped_column(ForeignKey(Agent.id), nullable=False)
    created_by: Mapped[Agent] = relationship(back_populates="created_tags")
    # `back_populates` is not really needed, but without it mypy terminates with an exception...


association_table = Table(
    "suppgram_conversation_tag_associations",
    Base.metadata,
    Column("conversation_id", ForeignKey("suppgram_conversations.id"), primary_key=True),
    Column("tag_id", ForeignKey("suppgram_tags.id"), primary_key=True),
)


class Conversation(Base):
    __tablename__ = "suppgram_conversations"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    customer_id: Mapped[int] = mapped_column(ForeignKey(Customer.id), nullable=False)
    customer: Mapped[Customer] = relationship(back_populates="conversations")
    tags: Mapped[List[Tag]] = relationship(secondary=association_table)
    assigned_workplace_id: Mapped[int] = mapped_column(ForeignKey(Workplace.id), nullable=True)
    assigned_workplace: Mapped[Workplace] = relationship(back_populates="conversations")
    state: Mapped[ConversationState] = mapped_column(Enum(ConversationState), nullable=False)
    messages: Mapped[List["ConversationMessage"]] = relationship(back_populates="conversation")
    customer_rating: Mapped[int] = mapped_column(Integer, nullable=True)


class ConversationMessage(Base):
    __tablename__ = "suppgram_conversation_messages"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    conversation_id: Mapped[int] = mapped_column(ForeignKey(Conversation.id), nullable=False)
    conversation: Mapped[Conversation] = relationship(back_populates="messages")
    kind: Mapped[MessageKind] = mapped_column(Enum(MessageKind), nullable=False)
    time_utc: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)
    text: Mapped[str] = mapped_column(String, nullable=True)


class Event(Base):
    __tablename__ = "suppgram_events"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    kind: Mapped[EventKind] = mapped_column(Enum(EventKind), nullable=False)
    time_utc: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)
    agent_id: Mapped[int] = mapped_column(ForeignKey(Agent.id), nullable=True)
    conversation_id: Mapped[int] = mapped_column(ForeignKey(Conversation.id), nullable=True)
    customer_id: Mapped[int] = mapped_column(ForeignKey(Customer.id), nullable=True)
    message_kind: Mapped[MessageKind] = mapped_column(Enum(MessageKind), nullable=True)
    message_media_kind: Mapped[MessageMediaKind] = mapped_column(
        Enum(MessageMediaKind), nullable=True
    )
    tag_id: Mapped[int] = mapped_column(ForeignKey(Tag.id), nullable=True)
    workplace_id: Mapped[int] = mapped_column(ForeignKey(Workplace.id), nullable=True)


class Models:
    """Abstraction layer over SQLAlchemy models.

    Default models are declared in `suppgram.storages.mongodb.collections` package."""

    def __init__(
        self,
        engine: AsyncEngine,
        customer_model: Any = Customer,
        agent_model: Any = Agent,
        workplace_model: Any = Workplace,
        conversation_model: Any = Conversation,
        conversation_message_model: Any = ConversationMessage,
        tag_model: Any = Tag,
        conversation_tag_association_table: Optional[Table] = association_table,
        event_model: Any = Event,
    ):
        """
        Parameters:
            engine: asynchronous SQLAlchemy engine
            customer_model: SQLAlchemy model for customers
            agent_model: SQLAlchemy model for agents
            workplace_model: SQLAlchemy model for workplaces
            conversation_model: SQLAlchemy model for conversations
            conversation_message_model: SQLAlchemy model for messages
            tag_model: SQLAlchemy model for conversation tags
            event_model: SQLAlchemy model for events
        """
        self._engine = engine
        self.customer_model = customer_model
        self.agent_model = agent_model
        self.workplace_model = workplace_model
        self.conversation_model = conversation_model
        self.conversation_message_model = conversation_message_model
        self.tag_model = tag_model
        self.event_model = event_model
        self._conversation_tag_association_table = conversation_tag_association_table

    async def initialize(self):
        tables_to_create = [
            built_in_model_or_table
            if isinstance(built_in_model_or_table, Table)
            else built_in_model_or_table.__table__
            for model_or_table, built_in_model_or_table in [
                (self.customer_model, Customer),
                (self.agent_model, Agent),
                (self.workplace_model, Workplace),
                (self.conversation_model, Conversation),
                (self.conversation_message_model, ConversationMessage),
                (self.tag_model, Tag),
                (self.event_model, Event),
                (self._conversation_tag_association_table, association_table),
            ]
            if model_or_table is built_in_model_or_table
        ]
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all, tables=tables_to_create)

    def convert_to_customer_model(self, identification: CustomerIdentification) -> Any:
        if identification.id is not None:
            raise ValueError("can't create customer with predefined ID")
        return self.customer_model(
            telegram_user_id=identification.telegram_user_id,
            shell_uuid=identification.shell_uuid.hex if identification.shell_uuid else None,
            pubnub_user_id=identification.pubnub_user_id,
            pubnub_channel_id=identification.pubnub_channel_id,
        )

    def apply_diff_to_customer_model(self, model: Customer, diff: CustomerDiff):
        if diff.telegram_first_name is not None:
            model.telegram_first_name = diff.telegram_first_name
        if diff.telegram_last_name is not None:
            model.telegram_last_name = diff.telegram_last_name
        if diff.telegram_username is not None:
            model.telegram_username = diff.telegram_username

    def convert_from_customer_model(self, customer: Customer) -> CustomerInterface:
        return CustomerInterface(
            id=customer.id,
            telegram_user_id=customer.telegram_user_id,
            telegram_first_name=customer.telegram_first_name,
            telegram_last_name=customer.telegram_last_name,
            telegram_username=customer.telegram_username,
            shell_uuid=UUID(customer.shell_uuid) if customer.shell_uuid else None,
            pubnub_user_id=customer.pubnub_user_id,
            pubnub_channel_id=customer.pubnub_channel_id,
        )

    def convert_to_agent_model(self, identification: AgentIdentification) -> Any:
        if identification.id is not None:
            raise ValueError("can't create agent with predefined ID")
        return self.agent_model(
            telegram_user_id=identification.telegram_user_id,
        )

    def apply_diff_to_agent_model(self, model: Agent, diff: AgentDiff):
        if diff.deactivated is not None:
            model.deactivated = diff.deactivated
        if diff.telegram_first_name is not None:
            model.telegram_first_name = diff.telegram_first_name
        if diff.telegram_last_name is not None:
            model.telegram_last_name = diff.telegram_last_name
        if diff.telegram_username is not None:
            model.telegram_username = diff.telegram_username

    def convert_from_agent_model(self, agent: Agent) -> AgentInterface:
        return AgentInterface(
            id=agent.id,
            deactivated=agent.deactivated,
            telegram_user_id=agent.telegram_user_id,
            telegram_first_name=agent.telegram_first_name,
            telegram_last_name=agent.telegram_last_name,
            telegram_username=agent.telegram_username,
        )

    def convert_to_workplace_model(
        self, agent_id: Any, identification: WorkplaceIdentification
    ) -> Any:
        return self.workplace_model(
            agent_id=agent_id,
            telegram_bot_id=identification.telegram_bot_id,
        )

    def convert_from_workplace_model(
        self, agent: AgentInterface, workplace: Workplace
    ) -> WorkplaceInterface:
        return WorkplaceInterface(
            id=workplace.id,
            telegram_user_id=agent.telegram_user_id,
            telegram_bot_id=workplace.telegram_bot_id,
            agent=agent,
        )

    def convert_to_message_model(
        self, conversation_id: Any, message: ConversaionMessageInterface
    ) -> Any:
        return self.conversation_message_model(
            conversation_id=conversation_id,
            kind=message.kind,
            time_utc=message.time_utc,
            text=message.text,
        )

    def convert_from_message_model(
        self, message: ConversationMessage
    ) -> ConversaionMessageInterface:
        return ConversaionMessageInterface(
            kind=message.kind, time_utc=message.time_utc, text=message.text
        )

    def make_tag_model(self, name: str, created_by: AgentInterface) -> Tag:
        return Tag(
            name=name,  # type: ignore
            created_at_utc=datetime.now(timezone.utc),  # type: ignore
            created_by_id=created_by.id,  # type: ignore
        )

    def convert_from_tag_model(self, tag: Tag, created_by: AgentInterface) -> TagInterface:
        return TagInterface(
            id=tag.id,
            name=tag.name,
            created_at_utc=tag.created_at_utc,
            created_by=created_by,
        )

    def convert_from_conversation_model(
        self, conversation: Conversation, with_messages: bool
    ) -> ConversationInterface:
        assigned_agent: Optional[AgentInterface] = None
        assigned_workplace: Optional[WorkplaceInterface] = None
        if conversation.assigned_workplace:
            assigned_agent = self.convert_from_agent_model(conversation.assigned_workplace.agent)
            assigned_workplace = self.convert_from_workplace_model(
                assigned_agent, conversation.assigned_workplace
            )
        return ConversationInterface(
            id=conversation.id,
            state=conversation.state,
            customer=self.convert_from_customer_model(conversation.customer),
            tags=[
                self.convert_from_tag_model(tag, self.convert_from_agent_model(tag.created_by))
                for tag in conversation.tags
            ],
            assigned_agent=assigned_agent,
            assigned_workplace=assigned_workplace,
            messages=[self.convert_from_message_model(message) for message in conversation.messages]
            if with_messages
            else UnavailableList[ConversaionMessageInterface](),
            customer_rating=conversation.customer_rating,
        )

    def make_customer_filter(self, identification: CustomerIdentification) -> ColumnElement:
        model = self.customer_model

        if identification.id is not None:
            return model.id == identification.id

        if identification.telegram_user_id is not None:
            return model.telegram_user_id == identification.telegram_user_id

        if identification.pubnub_user_id is not None:
            return (model.pubnub_user_id == identification.pubnub_user_id) & (
                model.pubnub_channel_id == identification.pubnub_channel_id
            )

        if identification.shell_uuid is not None:
            return model.shell_uuid == str(identification.shell_uuid)

        raise ValueError(
            f"received customer identification {identification} without supported non-null fields"
        )

    def make_agent_filter(self, identification: AgentIdentification) -> ColumnElement:
        model = self.agent_model

        if identification.id is not None:
            return model.id == identification.id

        if identification.telegram_user_id is not None:
            return model.telegram_user_id == identification.telegram_user_id

        raise AgentEmptyIdentification(identification)

    def make_workplace_filter(self, identification: WorkplaceIdentification) -> ColumnElement:
        workplace_model = self.workplace_model
        agent_model = self.agent_model
        if identification.id is not None:
            return (workplace_model.id == identification.id) & (
                workplace_model.agent_id == agent_model.id
            )
        if identification.telegram_user_id is not None:
            return (
                (workplace_model.telegram_bot_id == identification.telegram_bot_id)
                & (workplace_model.agent_id == agent_model.id)
                & (agent_model.telegram_user_id == identification.telegram_user_id)
            )
        raise WorkplaceEmptyIdentification(identification)

    def make_agent_workplaces_filter(self, agent: AgentInterface) -> ColumnElement:
        return self.workplace_model.agent_id == agent.id

    def make_current_customer_conversation_filter(self, customer: CustomerInterface):
        return (self.conversation_model.customer_id == customer.id) & (
            ~self.conversation_model.state.in_(FINAL_STATES)
        )

    def make_workplace_conversation_filter(
        self, identification: WorkplaceIdentification
    ) -> ColumnElement:
        return self.make_workplace_filter(identification) & (
            self.conversation_model.assigned_workplace_id == Workplace.id
        )

    def make_customer_conversations_filter(
        self, identification: CustomerIdentification
    ) -> ColumnElement:
        return self.make_customer_filter(identification) & (
            self.conversation_model.customer_id == self.customer_model.id
        )

    def make_agent_conversations_filter(self, identification: AgentIdentification) -> ColumnElement:
        return (
            self.make_agent_filter(identification)
            & (self.conversation_model.assigned_workplace_id == self.workplace_model.id)
            & (self.workplace_model.agent_id == self.agent_model.id)
        )

    def make_conversations_filter(
        self, conversation_ids: List[Any], unassigned_only: bool = False
    ) -> ColumnElement:
        result = self.conversation_model.id.in_(conversation_ids)
        if unassigned_only:
            result = result & (self.conversation_model.assigned_workplace_id == None)  # noqa
        return result

    def make_conversation_options(self, with_messages: bool) -> List[LoaderOption]:
        result: List[LoaderOption] = [
            joinedload(self.conversation_model.customer),
            selectinload(self.conversation_model.tags).joinedload(self.tag_model.created_by),
            joinedload(self.conversation_model.assigned_workplace).joinedload(
                self.workplace_model.agent
            ),
        ]
        if with_messages:
            result.append(selectinload(self.conversation_model.messages))
        return result

    def make_conversation_update_values(self, diff: ConversationDiff) -> Mapping[str, Any]:
        result: Dict[str, Any] = {}

        if diff.state is not None:
            result["state"] = diff.state

        if diff.assigned_workplace_id is SetNone:
            result["assigned_workplace_id"] = None
        elif diff.assigned_workplace_id is not None:
            result["assigned_workplace_id"] = diff.assigned_workplace_id

        if diff.customer_rating is not None:
            result["customer_rating"] = diff.customer_rating

        return result

    def convert_to_event_model(self, event: EventInterface) -> Event:
        return self.event_model(
            kind=event.kind,
            time_utc=event.time_utc.astimezone(timezone.utc),
            agent_id=event.agent_id,
            conversation_id=event.conversation_id,
            customer_id=event.customer_id,
            message_kind=event.message_kind,
            message_media_kind=event.message_media_kind,
            tag_id=event.tag_id,
            workplace_id=event.workplace_id,
        )

    def convert_from_event_model(self, event: Event) -> EventInterface:
        return EventInterface(
            kind=event.kind,
            time_utc=event.time_utc.replace(tzinfo=timezone.utc),
            agent_id=event.agent_id,
            conversation_id=event.conversation_id,
            customer_id=event.customer_id,
            message_kind=event.message_kind,
            message_media_kind=event.message_media_kind,
            tag_id=event.tag_id,
            workplace_id=event.workplace_id,
        )



import logging
from importlib import import_module
from typing import Optional, List, Any, Iterable, cast, Union
from uuid import UUID, uuid4

from suppgram.analytics import Reporter
from suppgram.backend import WorkplaceManager, Backend
from suppgram.entities import AgentIdentification
from suppgram.errors import NoStorageSpecified, NoFrontendSpecified
from suppgram.frontend import ManagerFrontend, CustomerFrontend, AgentFrontend
from suppgram.helpers import flat_gather
from suppgram.storage import Storage
from suppgram.texts.en import EnglishTextProvider
from suppgram.texts.interface import TextProvider

logger = logging.getLogger(__name__)


class Builder:
    """Provides a simple interface for building functional
    Suppgram application from separate components."""

    def __init__(self) -> None:
        # Implementation-specific objects are declared as `Any` to
        # avoid importing packages which may have missing dependencies.

        self._initialized = False

        self._storage: Optional[Storage] = None
        self._sqlalchemy_engine: Optional[Any] = None
        self._mongodb_client: Optional[Any] = None
        self._mongodb_database: Optional[Any] = None

        self._texts: Optional[TextProvider] = None

        # Helper classes, including implementation-specific:
        self._telegram_app_manager: Optional[Any] = None
        self._telegram_helper: Optional[Any] = None
        self._telegram_storage: Optional[Any] = None
        self._pubnub_configuration: Optional[Any] = None
        self._pubnub_message_converter: Optional[Any] = None
        self._workplace_managers: List[WorkplaceManager] = []

        self._backend: Optional[Backend] = None

        self._manager_frontend: Optional[ManagerFrontend] = None
        self._telegram_manager_bot_token: Optional[str] = None
        self._telegram_owner_id: Optional[int] = None

        self._customer_frontends: List[CustomerFrontend] = []
        self._telegram_customer_bot_token: Optional[str] = None
        self._shell_customer_uuid: Optional[UUID] = None
        self._pubnub_channel_group: Optional[str] = None

        self._agent_frontends: List[AgentFrontend] = []
        self._telegram_agent_bot_tokens: List[str] = []

    def build(self) -> "Builder":
        """Instantiate all configured components and raise exception if misconfigured."""
        self._build_backend()
        self._build_manager_frontend()
        self._build_customer_frontends()
        self._build_agent_frontends()
        from suppgram.backends.local import LocalBackend

        if (
            not self._customer_frontends
            and not self._agent_frontends
            and self._manager_frontend is None
            and isinstance(self._backend, LocalBackend)
        ):
            raise NoFrontendSpecified()

        return self

    async def start(self):
        """Start all configured components."""
        await self._initialize()
        await flat_gather(runnable.start() for runnable in self._get_runnables())

    def with_storage(self, storage: Storage) -> "Builder":
        """Configure arbitrary [Storage][suppgram.storage.Storage] instance."""
        if self._storage is not None:
            raise ValueError(
                f"can't use {type(storage).__name__} — "
                f"already instantiated {type(self._storage).__name__}"
            )

        self._storage = storage
        return self

    def with_sqlalchemy_storage(self, sqlalchemy_uri: str) -> "Builder":
        """Configure [SQLAlchemyStorage][suppgram.storages.sqlalchemy.SQLAlchemyStorage]."""
        if self._storage is not None:
            raise ValueError(
                f"can't use SQLAlchemy storage — already instantiated {type(self._storage).__name__}"
            )

        from suppgram.storages.sqlalchemy.models import Models
        from suppgram.storages.sqlalchemy.storage import SQLAlchemyStorage

        from sqlalchemy.ext.asyncio import create_async_engine

        self._sqlalchemy_engine = create_async_engine(sqlalchemy_uri)
        sqlalchemy_models = Models(engine=self._sqlalchemy_engine)
        self._storage = SQLAlchemyStorage(
            engine=self._sqlalchemy_engine,
            models=sqlalchemy_models,
        )
        return self

    def with_mongodb_storage(
        self, mongodb_uri: str, mongodb_database_name: Optional[str]
    ) -> "Builder":
        """Configure [MongoDBStorage][suppgram.storages.mongodb.MongoDBStorage]."""
        if self._storage is not None:
            raise ValueError(
                f"can't use MongoDB storage — already instantiated {type(self._storage).__name__}"
            )

        from motor.core import AgnosticDatabase
        from motor.motor_asyncio import AsyncIOMotorClient
        from suppgram.storages.mongodb.collections import Collections
        from suppgram.storages.mongodb.storage import MongoDBStorage

        self._mongodb_client = AsyncIOMotorClient(mongodb_uri, uuidRepresentation="standard")
        self._mongodb_database = cast(
            AgnosticDatabase,
            self._mongodb_client.get_default_database()
            if mongodb_database_name is None
            else self._mongodb_client.get_database(mongodb_database_name),
        )
        mongodb_collections = Collections(database=self._mongodb_database)
        self._storage = MongoDBStorage(collections=mongodb_collections)
        return self

    def with_texts(self, texts: TextProvider) -> "Builder":
        """Configure arbitrary [TextProvider][suppgram.texts.TextProvider] instance."""
        if self._texts is not None:
            raise ValueError(
                f"can't use {type(texts).__name__} — already instantiated {type(self._texts).__name__}"
            )

        self._texts = texts
        return self

    def with_texts_class_path(self, texts_class_path: str) -> "Builder":
        """Create [TextProvider][suppgram.texts.TextProvider] instance of given class.

        Assumes that its `__init__` method doesn't require any arguments."""
        if self._texts is not None:
            raise ValueError(
                f"can't use {texts_class_path} — already instantiated {type(self._texts).__name__}"
            )

        texts_module_name, texts_class_name = texts_class_path.rsplit(".", 1)
        texts_class = getattr(import_module(texts_module_name), texts_class_name)
        self._texts = texts_class()
        return self

    def with_telegram_manager_frontend(
        self, telegram_manager_bot_token: str, telegram_owner_id: Optional[int] = None
    ) -> "Builder":
        """
        Configure Telegram manager frontend.

        Arguments:
            telegram_manager_bot_token: Telegram bot token for manager bot
            telegram_owner_id: Telegram user ID of system administrator/owner
        """
        self._telegram_manager_bot_token = telegram_manager_bot_token
        self._telegram_owner_id = telegram_owner_id
        return self

    def with_telegram_customer_frontend(self, telegram_customer_bot_token: str) -> "Builder":
        """
        Configure Telegram customer frontend.

        Arguments:
            telegram_customer_bot_token: Telegram bot token for customer bot
        """
        self._telegram_customer_bot_token = telegram_customer_bot_token
        return self

    def with_shell_customer_frontend(self, uuid: Optional[UUID] = None) -> "Builder":
        """Configure shell customer frontend. Allows to chat with an agent directly in the shell.
        Useful for debug purposes."""
        if uuid is None:
            uuid = uuid4()
        self._shell_customer_uuid = uuid
        return self

    def with_pubnub_customer_frontend(
        self,
        pubnub_user_id: str,
        pubnub_channel_group: str,
        pubnub_message_converter_class_path: str = "suppgram.frontends.pubnub.DefaultMessageConverter",
    ) -> "Builder":
        """
        Configure PubNub customer frontend.

        Arguments:
            pubnub_user_id: PubNub user ID of the support user, on whose behalf agent messages will be sent to users
            pubnub_channel_group: ID of Pubnub channel group which includes all customers' chats with support
            pubnub_message_converter_class_path: allows to customize conversion between
                                                 Suppgram dataclasses and PubNub message JSONs
        """
        from suppgram.frontends.pubnub.configuration import make_pubnub_configuration
        from suppgram.frontends.pubnub.converter import make_pubnub_message_converter

        configuration = make_pubnub_configuration(pubnub_user_id)
        converter = make_pubnub_message_converter(pubnub_message_converter_class_path)

        self._pubnub_configuration = configuration
        self._pubnub_message_converter = converter
        self._pubnub_channel_group = pubnub_channel_group
        return self

    def with_telegram_agent_frontend(self, telegram_agent_bot_tokens: List[str]) -> "Builder":
        """
        Configure Telegram agent frontend.

        Arguments:
            telegram_agent_bot_tokens: list of Telegram bot tokens for agent bots. More tokens —
                                       more simultaneous chats with customers per agent
        """
        self._telegram_agent_bot_tokens = telegram_agent_bot_tokens
        return self

    def _build_storage(self) -> Storage:
        if self._storage is None:
            raise NoStorageSpecified()
        return self._storage

    def _build_texts(self) -> TextProvider:
        if self._texts is None:
            self._texts = EnglishTextProvider()
        return self._texts

    def _build_telegram_app_manager(self) -> Any:
        if self._telegram_app_manager is not None:
            return self._telegram_app_manager

        from suppgram.frontends.telegram.app_manager import TelegramAppManager

        tokens = [
            self._telegram_customer_bot_token,
            self._telegram_manager_bot_token,
            *self._telegram_agent_bot_tokens,
        ]
        self._telegram_app_manager = TelegramAppManager.from_tokens(tokens=[*filter(None, tokens)])
        return self._telegram_app_manager

    def _build_telegram_helper(self) -> Any:
        if self._telegram_helper is not None:
            return self._telegram_helper
        from suppgram.frontends.telegram.helper import TelegramHelper

        self._telegram_helper = TelegramHelper(
            manager_bot_token=self._telegram_manager_bot_token,
            app_manager=self._build_telegram_app_manager(),
            storage=self._build_telegram_storage(),
        )
        return self._telegram_helper

    def _build_telegram_storage(self) -> Any:
        if self._telegram_storage is not None:
            return self._telegram_storage
        self._build_storage()
        if self._sqlalchemy_engine is not None:
            from suppgram.bridges.sqlalchemy_telegram import SQLAlchemyTelegramBridge

            self._telegram_storage = SQLAlchemyTelegramBridge(self._sqlalchemy_engine)
        elif self._mongodb_database is not None:
            from suppgram.bridges.mongodb_telegram import MongoDBTelegramBridge

            self._telegram_storage = MongoDBTelegramBridge(self._mongodb_database)
        else:
            raise NoStorageSpecified()
        return self._telegram_storage

    def _build_workplace_managers(self) -> List[WorkplaceManager]:
        if self._workplace_managers:
            return self._workplace_managers
        if self._telegram_agent_bot_tokens:
            from suppgram.frontends.telegram.workplace_manager import (
                TelegramWorkplaceManager,
            )

            self._workplace_managers.append(
                TelegramWorkplaceManager(
                    self._telegram_agent_bot_tokens, self._build_telegram_app_manager()
                )
            )
        return self._workplace_managers

    def _build_backend(self) -> Backend:
        if self._backend is not None:
            return self._backend
        from suppgram.backends.local import LocalBackend

        self._backend = LocalBackend(
            storage=self._build_storage(),
            workplace_managers=self._build_workplace_managers(),
            texts=self._build_texts(),
        )
        return self._backend

    def _build_manager_frontend(self) -> Optional[ManagerFrontend]:
        if self._manager_frontend is not None:
            return self._manager_frontend

        if self._telegram_manager_bot_token:
            from suppgram.frontends.telegram.manager_frontend import (
                TelegramManagerFrontend,
            )

            self._manager_frontend = TelegramManagerFrontend(
                token=self._telegram_manager_bot_token,
                app_manager=self._build_telegram_app_manager(),
                backend=self._build_backend(),
                helper=self._build_telegram_helper(),
                storage=self._build_telegram_storage(),
                reporter=Reporter(self._build_storage()),
                texts=self._build_texts(),
            )

        return self._manager_frontend

    def _build_customer_frontends(self) -> List[CustomerFrontend]:
        if self._customer_frontends:
            return self._customer_frontends

        if self._telegram_customer_bot_token:
            from suppgram.frontends.telegram.customer_frontend import (
                TelegramCustomerFrontend,
            )

            logger.info("Initializing Telegram customer frontend")
            self._customer_frontends.append(
                TelegramCustomerFrontend(
                    token=self._telegram_customer_bot_token,
                    app_manager=self._build_telegram_app_manager(),
                    backend=self._build_backend(),
                    storage=self._build_telegram_storage(),
                    texts=self._build_texts(),
                )
            )

        if self._shell_customer_uuid:
            from suppgram.frontends.shell.customer_frontend import ShellCustomerFrontend

            logger.info("Initializing shell customer frontend")
            self._customer_frontends.append(
                ShellCustomerFrontend(backend=self._build_backend(), texts=self._build_texts())
            )

        if self._pubnub_configuration:
            from suppgram.frontends.pubnub.converter import MessageConverter
            from suppgram.frontends.pubnub.customer_frontend import (
                PubNubCustomerFrontend,
            )

            logger.info("Initializing PubNub customer frontend")

            self._customer_frontends.append(
                PubNubCustomerFrontend(
                    backend=self._build_backend(),
                    message_converter=cast(MessageConverter, self._pubnub_message_converter),
                    pubnub_channel_group=cast(str, self._pubnub_channel_group),
                    pubnub_configuration=self._pubnub_configuration,
                )
            )

        return self._customer_frontends

    def _build_agent_frontends(self) -> List[AgentFrontend]:
        if self._agent_frontends:
            return self._agent_frontends

        if self._telegram_agent_bot_tokens:
            from suppgram.frontends.telegram.agent_frontend import TelegramAgentFrontend

            self._agent_frontends.append(
                TelegramAgentFrontend(
                    agent_bot_tokens=self._telegram_agent_bot_tokens,
                    manager_bot_token=self._telegram_manager_bot_token,
                    app_manager=self._build_telegram_app_manager(),
                    backend=self._build_backend(),
                    helper=self._build_telegram_helper(),
                    storage=self._build_telegram_storage(),
                    texts=self._build_texts(),
                )
            )

        return self._agent_frontends

    async def _initialize(self):
        if self._initialized:
            return

        self.build()

        await self._storage.initialize()
        if self._telegram_storage:
            await self._telegram_storage.initialize()

        await flat_gather(runnable.initialize() for runnable in self._get_runnables())

        if self._telegram_owner_id:
            await self._backend.create_or_update_agent(
                AgentIdentification(telegram_user_id=self._telegram_owner_id)
            )

        self._initialized = True

    def _get_runnables(
        self,
    ) -> Iterable[Union[ManagerFrontend, CustomerFrontend, AgentFrontend]]:
        if manager_frontend := self._build_manager_frontend():
            yield manager_frontend
        yield from self._build_customer_frontends()
        yield from self._build_agent_frontends()



from typing import Any, List, Optional, Mapping, AsyncIterator

from pymongo import ReturnDocument
from pymongo.errors import DuplicateKeyError

from suppgram.entities import (
    Conversation,
    Message,
    WorkplaceIdentification,
    ConversationDiff,
    Customer,
    Tag,
    Agent,
    Workplace,
    AgentIdentification,
    AgentDiff,
    CustomerIdentification,
    CustomerDiff,
    Event,
)
from suppgram.errors import (
    AgentNotFound,
    WorkplaceNotFound,
    ConversationNotFound,
    TagAlreadyExists,
    ConversationAlreadyAssigned,
)
from suppgram.storage import Storage
from suppgram.storages.mongodb.collections import Collections, Document


class MongoDBStorage(Storage):
    """Implementation of [Storage][suppgram.storage.Storage] for MongoDB
    via [Motor](https://motor.readthedocs.io/) library."""

    def __init__(self, collections: Collections):
        """
        Parameters:
            collections: object storing collection names and conversion routines. Allows using custom
                         collection names, can be subclassed to customize BSON documents structure.
        """
        self._collections = collections

    async def create_or_update_customer(
        self, identification: CustomerIdentification, diff: Optional[CustomerDiff] = None
    ) -> Customer:
        filter_ = self._collections.make_customer_filter(identification)
        update = self._collections.make_customer_update(identification, diff)
        doc = await self._collections.customer_collection.find_one_and_update(
            filter_, update, upsert=identification.id is None, return_document=ReturnDocument.AFTER
        )
        if doc is None:
            raise ValueError("can't create customer with predefined ID")
        return self._collections.convert_to_customer(doc)

    async def find_customers_by_ids(self, customer_ids: List[Any]) -> List[Customer]:
        filter_ = self._collections.make_customers_filter(customer_ids)
        docs = await self._collections.customer_collection.find(filter_).to_list(None)
        return [self._collections.convert_to_customer(doc) for doc in docs]

    async def get_agent(self, identification: AgentIdentification) -> Agent:
        filter_ = self._collections.make_agent_filter(identification)
        doc = await self._collections.agent_collection.find_one(filter_)
        if doc is None:
            raise AgentNotFound(identification)
        return self._collections.convert_to_agent(doc)

    async def find_all_agents(self) -> AsyncIterator[Agent]:
        docs = self._collections.agent_collection.find({})
        async for doc in docs:
            yield self._collections.convert_to_agent(doc)

    async def create_or_update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff] = None
    ) -> Agent:
        try:
            return await self._update_agent(identification, diff, upsert=identification.id is None)
        except AgentNotFound:
            raise ValueError("can't create agent with predefined ID")

    async def update_agent(self, identification: AgentIdentification, diff: AgentDiff) -> Agent:
        return await self._update_agent(identification, diff, upsert=False)

    async def _update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff], upsert: bool
    ) -> Agent:
        filter_ = self._collections.make_agent_filter(identification)
        update = self._collections.convert_to_agent_update(identification, diff)
        doc = await self._collections.agent_collection.find_one_and_update(
            filter_, update, upsert=upsert, return_document=ReturnDocument.AFTER
        )
        if doc is None:
            raise AgentNotFound(identification)
        return self._collections.convert_to_agent(doc)

    async def get_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        filter_ = self._collections.make_agent_filter(identification)
        agent_doc = await self._collections.agent_collection.find_one(filter_)
        if agent_doc is None:
            raise WorkplaceNotFound(identification)
        return self._collections.convert_to_workplace(identification, agent_doc)

    async def find_workplaces_by_ids(self, workplace_ids: List[Any]) -> List[Workplace]:
        workplace_ids_set = set(workplace_ids)
        filter_ = self._collections.make_agents_filter_by_workplace_ids(workplace_ids)
        agent_docs = await self._collections.agent_collection.find(filter_).to_list(None)
        # TODO Agents here will be copied as many times as many of their
        #      workplaces are fetched. Might optimize a bit if got spare time.
        return [
            workplace
            for agent_doc in agent_docs
            for workplace in self._collections.convert_to_workplaces(agent_doc)
            if workplace.id in workplace_ids_set
        ]

    async def get_agent_workplaces(self, agent: Agent) -> List[Workplace]:
        filter_ = self._collections.make_agent_filter(agent.identification)
        agent_doc = await self._collections.agent_collection.find_one(filter_)
        if agent_doc is None:
            raise AgentNotFound(agent.identification)
        return self._collections.convert_to_workplaces(agent_doc)

    async def get_or_create_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        filter_ = self._collections.make_agent_filter(identification)
        agent_doc = await self._collections.agent_collection.find_one(filter_)
        if agent_doc is None:
            raise AgentNotFound(identification.to_agent_identification())
        if identification.id is None:
            agent_id = agent_doc["_id"]
            update = self._collections.convert_to_workspace_update(agent_id, identification)
            agent_doc = await self._collections.agent_collection.find_one_and_update(
                filter_, update, return_document=ReturnDocument.AFTER
            )
        return self._collections.convert_to_workplace(identification, agent_doc)

    async def create_tag(self, name: str, created_by: Agent) -> Tag:
        try:
            doc = self._collections.convert_to_tag_document(name, created_by)
            await self._collections.tag_collection.insert_one(doc)
            return self._collections.convert_to_tag(doc, {created_by.id: created_by})
        except DuplicateKeyError as exc:
            raise TagAlreadyExists(name) from exc

    async def find_all_tags(self) -> List[Tag]:
        docs = await self._collections.tag_collection.find({}).to_list(None)
        agent_ids = self._collections.extract_agent_ids(docs)
        filter_ = self._collections.make_agents_filter(agent_ids)
        agents = [
            self._collections.convert_to_agent(doc)
            for doc in await self._collections.agent_collection.find(filter_).to_list(None)
        ]
        agent_by_id = {agent.id: agent for agent in agents}
        return [self._collections.convert_to_tag(doc, agent_by_id) for doc in docs]

    async def get_or_create_conversation(self, customer: Customer) -> Conversation:
        filter_ = self._collections.make_conversation_filter_by_customer(customer)
        update = self._collections.make_conversation_update(customer=customer)
        doc = await self._collections.conversation_collection.find_one_and_update(
            filter_,
            update,
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        return await self._convert_single_conversation(doc)

    async def find_conversations_by_ids(
        self, conversation_ids: List[Any], with_messages: bool = False
    ) -> List[Conversation]:
        filter_ = self._collections.make_conversations_filter(conversation_ids)
        projection = self._collections.make_conversation_projection(with_messages=with_messages)
        docs = await self._collections.conversation_collection.find(
            filter_, projection=projection
        ).to_list(None)
        return await self._convert_multiple_conversations(docs)

    async def find_all_conversations(
        self, with_messages: bool = False
    ) -> AsyncIterator[Conversation]:
        projection = self._collections.make_conversation_projection(with_messages=with_messages)
        docs = self._collections.conversation_collection.find({}, projection=projection)
        batch: List[Document] = []
        async for doc in docs:
            batch.append(doc)
            if len(batch) >= 100:
                for conv in await self._convert_multiple_conversations(batch):
                    yield conv
                batch.clear()
        for conv in await self._convert_multiple_conversations(batch):
            yield conv

    async def count_all_conversations(self) -> int:
        return await self._collections.conversation_collection.count_documents({})

    async def update_conversation(
        self, id: Any, diff: ConversationDiff, unassigned_only: bool = False
    ):
        filter_ = self._collections.make_conversation_filter(id, unassigned_only=False)
        doc = await self._collections.conversation_collection.find_one(filter_)
        if doc is None:
            raise ConversationNotFound()

        workplaces: Mapping[Any, Workplace] = {}
        if (identification := diff.assigned_workplace_identification) is not None:
            workplace = await self.get_workplace(identification)
            workplaces = {workplace.id: workplace}
        filter_ = self._collections.make_conversation_filter(id, unassigned_only=unassigned_only)
        update = self._collections.make_conversation_update(diff=diff, workplaces=workplaces)
        result = await self._collections.conversation_collection.update_one(filter_, update)
        if result.matched_count == 0:
            raise ConversationAlreadyAssigned()

    async def get_agent_conversation(self, identification: WorkplaceIdentification) -> Conversation:
        workplace_id = identification.id
        if workplace_id is None:
            workplace = await self.get_workplace(identification)
            workplace_id = workplace.id
        filter_ = self._collections.make_workplace_conversation_filter(workplace_id)
        doc = await self._collections.conversation_collection.find_one(filter_)
        if doc is None:
            raise ConversationNotFound()
        return await self._convert_single_conversation(doc)

    async def find_customer_conversations(
        self, customer: Customer, with_messages: bool = False
    ) -> List[Conversation]:
        filter_ = self._collections.make_customer_conversations_filter(customer.id)
        projection = self._collections.make_conversation_projection(with_messages=with_messages)
        docs = await self._collections.conversation_collection.find(
            filter_, projection=projection
        ).to_list(None)
        return await self._convert_multiple_conversations(docs, customers={customer.id: customer})

    async def find_agent_conversations(
        self, agent: Agent, with_messages: bool = False
    ) -> List[Conversation]:
        filter_ = self._collections.make_agent_conversations_filter(agent.id)
        projection = self._collections.make_conversation_projection(with_messages=with_messages)
        docs = await self._collections.conversation_collection.find(
            filter_, projection=projection
        ).to_list(None)
        return await self._convert_multiple_conversations(docs)

    async def _convert_single_conversation(self, conv_doc: Document) -> Conversation:
        related_ids = self._collections.extract_conversation_related_ids(conv_doc)
        customer = await self.create_or_update_customer(related_ids.customer_identification)
        customers = {customer.id: customer}
        workplaces: Mapping[Any, Workplace] = {}
        if (workplace_identification := related_ids.assigned_workplace_identification) is not None:
            assigned_workplace = await self.get_workplace(workplace_identification)
            workplaces = {assigned_workplace.id: assigned_workplace}
        tags = {tag.id: tag for tag in await self.find_all_tags()}
        return self._collections.convert_to_conversation(conv_doc, customers, workplaces, tags)

    async def _convert_multiple_conversations(
        self, conv_docs: List[Document], customers: Mapping[Any, Customer] = {}
    ) -> List[Conversation]:
        related_ids = [self._collections.extract_conversation_related_ids(doc) for doc in conv_docs]
        if not customers:
            customer_ids = [r.customer_id for r in related_ids]
            customers = {
                customer.id: customer for customer in await self.find_customers_by_ids(customer_ids)
            }
        workplace_ids = [
            r.assigned_workplace_id for r in related_ids if r.assigned_workplace_id is not None
        ]
        workplaces = {
            workplace.id: workplace
            for workplace in await self.find_workplaces_by_ids(workplace_ids)
        }
        tags = {tag.id: tag for tag in await self.find_all_tags()}
        return [
            self._collections.convert_to_conversation(doc, customers, workplaces, tags)
            for doc in conv_docs
        ]

    async def save_message(self, conversation: Conversation, message: Message):
        filter_ = self._collections.make_conversation_filter(conversation.id, unassigned_only=False)
        update = self._collections.make_message_update(message)
        result = await self._collections.conversation_collection.update_one(filter_, update)
        if result.matched_count == 0:
            raise ConversationNotFound()

    async def save_event(self, event: Event):
        doc = self._collections.convert_to_event_document(event)
        await self._collections.event_collection.insert_one(doc)

    async def find_all_events(self) -> AsyncIterator[Event]:
        docs = self._collections.event_collection.find({})
        async for doc in docs:
            yield self._collections.convert_to_event(doc)

    async def count_all_events(self) -> int:
        return await self._collections.event_collection.count_documents({})



from typing import (
    List,
    Any,
    TypeVar,
    Optional,
    AsyncIterator,
)

from sqlalchemy import (
    select,
    update,
    Column,
)
from sqlalchemy.exc import NoResultFound, IntegrityError
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker
from sqlalchemy.orm import (
    joinedload,
)
from sqlalchemy.sql.functions import count

from suppgram.entities import (
    Agent,
    AgentDiff,
    AgentIdentification,
    Conversation,
    ConversationDiff,
    ConversationState,
    Tag,
    Customer,
    CustomerDiff,
    CustomerIdentification,
    Message,
    Workplace,
    WorkplaceIdentification,
    Event,
)
from suppgram.errors import (
    ConversationNotFound,
    WorkplaceNotFound,
    AgentNotFound,
    ConversationAlreadyAssigned,
    TagAlreadyExists,
)
from suppgram.storage import Storage
from suppgram.storages.sqlalchemy.models import (
    Base,
    association_table,
    Models,
)

T = TypeVar("T", bound=Base)


class SQLAlchemyStorage(Storage):
    """Implementation of [Storage][suppgram.storage.Storage] for SQLAlchemy."""

    def __init__(self, engine: AsyncEngine, models: Models):
        """
        Parameters:
            engine: asynchronous SQLAlchemy engine
            models: collection of SQLAlchemy model types.
                    Allows using custom models, possibly enriched with some
                    business-specific fields or stripped of unnecessary columns
                    (e.g. for frontends you are not going to use)."""
        self._engine = engine
        self._models = models
        self._session = async_sessionmaker(bind=engine)

    async def initialize(self):
        await super().initialize()
        await self._models.initialize()

    async def create_or_update_customer(
        self,
        identification: CustomerIdentification,
        diff: Optional[CustomerDiff] = None,
    ) -> Customer:
        filter_ = self._models.make_customer_filter(identification)
        async with self._session() as session, session.begin():
            select_query = select(self._models.customer_model).where(filter_).with_for_update()
            customer = (await session.execute(select_query)).scalars().one_or_none()
            if customer is None:
                customer = self._models.convert_to_customer_model(identification)
                if diff is not None:
                    self._models.apply_diff_to_customer_model(customer, diff)
                session.add(customer)
                await session.flush()
            elif diff is not None:
                self._models.apply_diff_to_customer_model(customer, diff)
                session.add(customer)
            return self._models.convert_from_customer_model(customer)

    async def get_agent(self, identification: AgentIdentification) -> Agent:
        async with self._session() as session:
            select_query = select(self._models.agent_model).where(
                self._models.make_agent_filter(identification)
            )
            agent = (await session.execute(select_query)).scalars().one_or_none()
            if agent is None:
                raise AgentNotFound(identification)
            return self._models.convert_from_agent_model(agent)

    async def find_all_agents(self) -> AsyncIterator[Agent]:
        async with self._session() as session:
            select_query = select(self._models.agent_model)
            async for agent in await session.stream_scalars(select_query):
                yield self._models.convert_from_agent_model(agent)

    async def create_or_update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff] = None
    ) -> Agent:
        filter_ = self._models.make_agent_filter(identification)
        async with self._session() as session, session.begin():
            select_query = select(self._models.agent_model).where(filter_).with_for_update()
            agent = (await session.execute(select_query)).scalars().one_or_none()
            if agent is None:
                agent = self._models.convert_to_agent_model(identification)
                if diff is not None:
                    self._models.apply_diff_to_agent_model(agent, diff)
                session.add(agent)
                await session.flush()
            elif diff is not None:
                self._models.apply_diff_to_agent_model(agent, diff)
                session.add(agent)
            return self._models.convert_from_agent_model(agent)

    async def update_agent(self, identification: AgentIdentification, diff: AgentDiff) -> Agent:
        filter_ = self._models.make_agent_filter(identification)
        async with self._session() as session, session.begin():
            select_query = select(self._models.agent_model).where(filter_).with_for_update()
            agent = (await session.execute(select_query)).scalars().one_or_none()
            if agent is None:
                raise AgentNotFound(identification)
            self._models.apply_diff_to_agent_model(agent, diff)
            session.add(agent)
            return self._models.convert_from_agent_model(agent)

    async def get_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        async with self._session() as session:
            select_query = select(self._models.workplace_model).where(
                self._models.make_workplace_filter(identification)
            )
            workplace = (await session.execute(select_query)).scalars().one_or_none()
            if workplace is None:
                raise WorkplaceNotFound(identification)
            return workplace

    async def get_agent_workplaces(self, agent: Agent) -> List[Workplace]:
        async with self._session() as session:
            select_query = select(self._models.workplace_model).where(
                self._models.make_agent_workplaces_filter(agent)
            )
            workplaces = (await session.execute(select_query)).scalars().all()
            return [
                self._models.convert_from_workplace_model(agent, workplace)
                for workplace in workplaces
            ]

    async def get_or_create_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        async with self._session() as session, session.begin():
            select_query = (
                select(self._models.workplace_model, self._models.agent_model)
                .options(joinedload(self._models.workplace_model.agent))
                .where(self._models.make_workplace_filter(identification))
            )
            workplace = (await session.execute(select_query)).scalars().one_or_none()
            if workplace is None:
                agent_identification = identification.to_agent_identification()
                select_query = select(self._models.agent_model).where(
                    self._models.make_agent_filter(agent_identification)
                )
                agent = (await session.execute(select_query)).scalars().one_or_none()
                if agent is None:
                    raise AgentNotFound(agent_identification)
                workplace = self._models.convert_to_workplace_model(agent.id, identification)
                session.add(workplace)
                await session.flush()
            else:
                agent = workplace.agent
            return self._models.convert_from_workplace_model(
                self._models.convert_from_agent_model(agent), workplace
            )

    async def create_tag(self, name: str, created_by: Agent) -> Tag:
        try:
            async with self._session() as session, session.begin():
                tag = self._models.make_tag_model(name, created_by)
                session.add(tag)
                await session.flush()
                return self._models.convert_from_tag_model(tag, created_by)
        except IntegrityError as exc:
            raise TagAlreadyExists(name) from exc

    async def find_all_tags(self) -> List[Tag]:
        async with self._session() as session:
            select_query = select(self._models.tag_model).options(
                joinedload(self._models.tag_model.created_by)
            )
            tags = (await session.execute(select_query)).scalars().all()
            return [
                self._models.convert_from_tag_model(
                    tag, self._models.convert_from_agent_model(tag.created_by)
                )
                for tag in tags
            ]

    async def get_or_create_conversation(self, customer: Customer) -> Conversation:
        async with self._session() as session, session.begin():
            options = self._models.make_conversation_options(with_messages=True)
            select_query = (
                select(self._models.conversation_model)
                .options(*options)
                .where(self._models.make_current_customer_conversation_filter(customer))
            )
            conv = (await session.execute(select_query)).scalars().one_or_none()
            assigned_agent: Optional[Agent] = None
            assigned_workplace: Optional[Workplace] = None
            if conv is None:
                conv = self._models.conversation_model(
                    customer_id=customer.id,
                    state=ConversationState.NEW,
                )
                session.add(conv)
                await session.flush()
                messages = []
                tags = []
            else:
                if conv.assigned_workplace:
                    assigned_agent = self._models.convert_from_agent_model(
                        conv.assigned_workplace.agent
                    )
                    assigned_workplace = self._models.convert_from_workplace_model(
                        assigned_agent, conv.assigned_workplace
                    )
                messages = [self._models.convert_from_message_model(msg) for msg in conv.messages]
                tags = [
                    self._models.convert_from_tag_model(
                        tag, self._models.convert_from_agent_model(tag.created_by)
                    )
                    for tag in conv.tags
                ]
            return Conversation(
                id=conv.id,
                state=conv.state,
                customer=customer,
                tags=tags,
                assigned_agent=assigned_agent,
                assigned_workplace=assigned_workplace,
                messages=messages,
            )

    async def find_conversations_by_ids(
        self, conversation_ids: List[Any], with_messages: bool = False
    ) -> List[Conversation]:
        async with self._session() as session:
            options = self._models.make_conversation_options(with_messages=with_messages)
            select_query = (
                select(self._models.conversation_model)
                .options(*options)
                .where(self._models.make_conversations_filter(conversation_ids))
            )
            convs = (await session.execute(select_query)).scalars().all()
            return [
                self._models.convert_from_conversation_model(conv, with_messages=with_messages)
                for conv in convs
            ]

    async def find_all_conversations(
        self, with_messages: bool = False
    ) -> AsyncIterator[Conversation]:
        async with self._session() as session:
            options = self._models.make_conversation_options(with_messages=with_messages)
            select_query = select(self._models.conversation_model).options(*options)
            async for conv in await session.stream_scalars(select_query):
                yield self._models.convert_from_conversation_model(
                    conv, with_messages=with_messages
                )

    async def count_all_conversations(self) -> int:
        async with self._session() as session:
            select_query = select(count(self._models.conversation_model.id))
            return (await session.execute(select_query)).scalar_one()

    async def update_conversation(
        self, id: Any, diff: ConversationDiff, unassigned_only: bool = False
    ):
        async with self._session() as session, session.begin():
            filter_ = self._models.make_conversations_filter([id])
            select_query = select(self._models.conversation_model).where(filter_)
            conv = (await session.execute(select_query)).scalars().one_or_none()
            if conv is None:
                raise ConversationNotFound()

            if update_values := self._models.make_conversation_update_values(diff):
                filter_ = self._models.make_conversations_filter(
                    [id], unassigned_only=unassigned_only
                )
                update_query = (
                    update(self._models.conversation_model).where(filter_).values(**update_values)
                )
                result = await session.execute(update_query)
                if unassigned_only and result.rowcount == 0:
                    raise ConversationAlreadyAssigned()

            if diff.added_tags:
                insert_query = association_table.insert().values(
                    [(conv.id, tag.id) for tag in diff.added_tags]
                )
                await session.execute(insert_query)

            if diff.removed_tags:
                filter_ = (Column("conversation_id") == conv.id) & (
                    Column("tag_id").in_(tag.id for tag in diff.removed_tags)
                )
                delete_query = association_table.delete().where(filter_)
                await session.execute(delete_query)

    async def get_agent_conversation(self, identification: WorkplaceIdentification) -> Conversation:
        try:
            async with self._session() as session:
                filter_ = self._models.make_workplace_conversation_filter(identification)
                options = self._models.make_conversation_options(with_messages=True)
                select_query = (
                    select(
                        self._models.conversation_model,
                        self._models.workplace_model,
                        self._models.agent_model,
                    )
                    .where(filter_)
                    .options(*options)
                )
                conv = (await session.execute(select_query)).scalars().one()
                return self._models.convert_from_conversation_model(conv, with_messages=True)
        except NoResultFound as exc:
            raise ConversationNotFound() from exc

    async def find_customer_conversations(
        self, customer: Customer, with_messages: bool = False
    ) -> List[Conversation]:
        async with self._session() as session:
            filter_ = self._models.make_customer_conversations_filter(customer.identification)
            options = self._models.make_conversation_options(with_messages=with_messages)
            select_query = (
                select(self._models.conversation_model, self._models.customer_model)
                .where(filter_)
                .options(*options)
            )
            convs = (await session.execute(select_query)).scalars().all()
            return [
                self._models.convert_from_conversation_model(conv, with_messages=with_messages)
                for conv in convs
            ]

    async def find_agent_conversations(
        self, agent: Agent, with_messages: bool = False
    ) -> List[Conversation]:
        async with self._session() as session:
            filter_ = self._models.make_agent_conversations_filter(agent.identification)
            options = self._models.make_conversation_options(with_messages=with_messages)
            select_query = (
                select(
                    self._models.conversation_model,
                    self._models.workplace_model,
                    self._models.agent_model,
                )
                .where(filter_)
                .options(*options)
            )
            convs = (await session.execute(select_query)).scalars().all()
            return [
                self._models.convert_from_conversation_model(conv, with_messages=with_messages)
                for conv in convs
            ]

    async def save_message(self, conversation: Conversation, message: Message):
        try:
            async with self._session() as session, session.begin():
                session.add(self._models.convert_to_message_model(conversation.id, message))
        except IntegrityError:
            raise ConversationNotFound()

    async def save_event(self, event: Event):
        async with self._session() as session, session.begin():
            session.add(self._models.convert_to_event_model(event))

    async def find_all_events(self) -> AsyncIterator[Event]:
        async with self._session() as session:
            select_query = select(self._models.event_model).order_by(self._models.event_model.id)
            async for event in await session.stream_scalars(select_query):
                yield self._models.convert_from_event_model(event)

    async def count_all_events(self) -> int:
        async with self._session() as session:
            select_query = select(count(self._models.event_model.id))
            return (await session.execute(select_query)).scalar_one()



from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Optional, Any, List, Union
from uuid import UUID


class _SetNone:
    pass


SetNone = _SetNone()


@dataclass(frozen=True)
class CustomerIdentification:
    """Subset of [Customer][suppgram.entities.Customer] fields allowing to uniquely identify the customer."""

    id: Optional[Any] = None
    telegram_user_id: Optional[int] = None
    shell_uuid: Optional[UUID] = None
    pubnub_user_id: Optional[str] = None
    pubnub_channel_id: Optional[str] = None


@dataclass(frozen=True)
class Customer:
    """
    Describes a particular customer who is interacting with the support system.

    Contains the data necessary to identify the external system customer communicates
    through (e. g. Telegram bot, Instagram business account, PubNub chat, etc.),
    the data necessary to identify the customer and send them messages within that system,
    and all available metadata (e.g. first name, last name) that can be useful.

    Attributes:
        id: internal ID of the customer. Type and format may depend on chosen [storage][suppgram.storage.Storage].
        telegram_user_id: Telegram user ID, if the customer is communicating via a Telegram bot.
        telegram_first_name: Telegram user first name, if the customer is communicating via a Telegram bot
                             and has non-empty first name.
        telegram_last_name: Telegram user last name, if the customer is communicating via a Telegram bot
                            and has non-empty last name.
        telegram_username: Telegram user username, if the customer is communicating via a Telegram bot
                           and their Telegram privacy settings allow access to their username.
        shell_uuid: UUID of the customer, if the customer is communicating via shell interface.
        pubnub_user_id: PubNub user ID, if the customer is communicating via a PubNub channel.
        pubnub_channel_id: PubNub channel ID, if the customer is communicating via a PubNub channel.
    """

    id: Any

    telegram_user_id: Optional[int] = None
    telegram_first_name: Optional[str] = None
    telegram_last_name: Optional[str] = None
    telegram_username: Optional[str] = None

    shell_uuid: Optional[UUID] = None

    pubnub_user_id: Optional[str] = None
    pubnub_channel_id: Optional[str] = None

    @property
    def identification(self) -> CustomerIdentification:
        return CustomerIdentification(id=self.id)


@dataclass(frozen=True)
class CustomerDiff:
    """Describes an update to a [Customer][suppgram.entities.Customer] object.

    Is, in fact, a subset of [Customer][suppgram.entities.Customer] fields representing optional metadata.
    """

    telegram_first_name: Optional[str] = None
    telegram_last_name: Optional[str] = None
    telegram_username: Optional[str] = None


@dataclass(frozen=True)
class AgentIdentification:
    """Subset of [Agent][suppgram.entities.Agent] fields allowing to uniquely identify the agent."""

    id: Optional[Any] = None
    telegram_user_id: Optional[int] = None


@dataclass(frozen=True)
class Agent:
    """
    Describes support agent who is able to communicate with customers via the support system or manage other agents.

    Attributes:
        id: internal ID of the agent. Type and format may depend on chosen [storage][suppgram.storage.Storage].
        telegram_user_id: Telegram user ID, if the agent is working via Telegram interface.
        telegram_first_name: Telegram user first name, if the agent is working via Telegram interface
                             and has non-empty first name.
        telegram_last_name: Telegram user last name, if the agent is working via Telegram interface
                            and has non-empty last name.
        telegram_username: Telegram user username, if the agent is working via Telegram interface
                           and their Telegram privacy settings allow access to their username.
    """

    id: Any
    deactivated: bool

    telegram_user_id: Optional[int] = None
    telegram_first_name: Optional[str] = None
    telegram_last_name: Optional[str] = None
    telegram_username: Optional[str] = None

    @property
    def identification(self) -> AgentIdentification:
        return AgentIdentification(id=self.id)


@dataclass(frozen=True)
class AgentDiff:
    """Describes an update to a [Agent][suppgram.entities.Agent] object.

    Is, in fact, a subset of [Agent][suppgram.entities.Agent] fields representing optional metadata.
    """

    deactivated: Optional[bool] = None

    telegram_first_name: Optional[str] = None
    telegram_last_name: Optional[str] = None
    telegram_username: Optional[str] = None


@dataclass(frozen=True)
class WorkplaceIdentification:
    """Subset of [Workplace][suppgram.entities.Workplace] fields allowing to uniquely identify the workplace."""

    id: Optional[Any] = None

    telegram_user_id: Optional[int] = None
    telegram_bot_id: Optional[int] = None

    def to_agent_identification(self) -> AgentIdentification:
        if self.telegram_user_id is not None:
            return AgentIdentification(telegram_user_id=self.telegram_user_id)
        raise RuntimeError(
            ".to_agent_identification() should not be called for already existing workplaces with IDs"
        )


@dataclass(frozen=True)
class Workplace:
    """
    Describes support agent's workplace.

    Workplace is an abstraction over a way agent conveys their messages.
    Examples of a workplace include:

      * private Telegram chat with one of the agent bots, identified by user ID and bot ID;
      * PubNub chat, identified by chat ID;
      * web interface session, identified by active websocket identifier.

    Attributes:
        id: internal ID of the workplace. Type and format may depend on chosen [storage][suppgram.storage.Storage].
        agent: agent the workplace belongs to.
    """

    id: Any

    telegram_user_id: Optional[int]
    telegram_bot_id: Optional[int]

    agent: Agent

    @property
    def identification(self) -> WorkplaceIdentification:
        return WorkplaceIdentification(
            id=self.id, telegram_user_id=self.telegram_user_id, telegram_bot_id=self.telegram_bot_id
        )


class MessageKind(str, Enum):
    """Enumeration describing [message][suppgram.entities.Message] kind."""

    FROM_CUSTOMER = "from_customer"
    """ Regular message from a customer. """
    FROM_AGENT = "from_agent"
    """ Regular message from a currently assigned agent. """
    POSTPONED = "postponed"
    """ Internal message marking the moment the conversation was postponed by the agent."""
    RESOLVED = "resolved"
    """ Internal message marking the moment the conversation was resolved by the agent. """


@dataclass(frozen=True)
class Message:
    """
    Describes message within a [conversation][suppgram.entities.Conversation].

    May be a message from customer, agent, or an internal message corresponding
    to an event (e.g. conversation resolution) depending on `kind`.

    Attributes:
        kind: message kind.
        time_utc: date and time of message creation.
        text: message text, if this is a text message.
    """

    kind: MessageKind
    time_utc: datetime
    text: Optional[str] = None
    # don't forget to add new `MessageMediaKind`
    # when implementing images, videos, etc.

    @property
    def media_kind(self) -> "MessageMediaKind":
        if self.text is not None:
            return MessageMediaKind.TEXT
        return MessageMediaKind.NONE


class ConversationState(str, Enum):
    """
    Enumeration describing current state of a [conversation][suppgram.entities.Conversation].

    Attributes:
        NEW: newly created conversation waiting for an agent to be assigned
        ASSIGNED: in-progress conversation with an assigned agent
        RESOLVED: conversation resolved by an agent
    """

    NEW = "new"
    ASSIGNED = "assigned"
    RESOLVED = "resolved"


FINAL_STATES = [ConversationState.RESOLVED]
""" List of states in which the conversation is not considered
 ready to accept new messages from the customer. """


@dataclass(frozen=True)
class Tag:
    """
    Describes tag that can be used to label a [conversation][suppgram.entities.Conversation].

    Attributes:
        id: internal ID of the tag. Type and format may depend on chosen [storage][suppgram.storage.Storage].
        name: tag name.
        created_at_utc: date and time when the tag was created.
        created_by: agent who has created the tag.
    """

    id: Any
    name: str
    created_at_utc: datetime
    created_by: Agent


@dataclass(frozen=True)
class Conversation:
    """
    Describes a particular conversation between customer and support agent(s).

    When a [customer][suppgram.entities.Customer] sends a message into the support system,
    a new conversation is created. It can then be assigned to an [agent][suppgram.entities.Agent]
    and resolved by them. If the customer sends a message again, previously resolved conversations
    are ignored and a new one is spawned.

    Attributes:
        id: internal ID of the conversation. Type and format may depend on chosen [storage][suppgram.storage.Storage].
        state: current state of the conversation.
        customer: customer who has started the conversation.
        tags: list of tags added to the conversation.
        assigned_agent: agent currently assigned to the conversation.
        assigned_workplace: workplace currently assigned to the conversation.
        messages: list of [messages][suppgram.entities.Message] from the beginning of the conversation.
                  Includes customer's messages, agent's messages, and internal messages marking particular
                  events (e.g. conversation resolution).
        customer_rating: 5-star rating given by the customer after the resolution of the conversation.
    """

    id: Any
    state: ConversationState
    customer: Customer
    tags: List[Tag]
    assigned_agent: Optional[Agent] = None
    assigned_workplace: Optional[Workplace] = None
    messages: List[Message] = field(default_factory=list)
    customer_rating: Optional[int] = None


@dataclass(frozen=True)
class ConversationDiff:
    """
    Describes an update to a [Conversation][suppgram.entities.Conversation] object.

    Attributes:
        state: new state of the conversation.
        assigned_workplace_id: ID of the [workplace][suppgram.entities.Workplace] newly assigned to the
                               conversation. Note that workplaces have many-to-one relationship with
                               [agents][suppgram.entities.Agent], thus agent ID is not necessary here.
        added_tags: list of tags to add to the conversation.
        removed_tags: list of tags to remove from the conversation.
        customer_rating: 5-star rating newly given by a customer.
    """

    state: Optional[ConversationState] = None
    assigned_workplace_id: Union[Optional[Any], _SetNone] = None
    added_tags: Optional[List[Tag]] = None
    removed_tags: Optional[List[Tag]] = None
    customer_rating: Optional[int] = None

    @property
    def assigned_workplace_identification(self) -> Optional[WorkplaceIdentification]:
        if self.assigned_workplace_id in (None, SetNone):
            return None
        return WorkplaceIdentification(id=self.assigned_workplace_id)


# Data classes for analytics


class EventKind(str, Enum):
    AGENT_ASSIGNED = "agent_assigned"
    CONVERSATION_POSTPONED = "conversation_postponed"
    CONVERSATION_RATED = "conversation_rated"
    CONVERSATION_RESOLVED = "conversation_resolved"
    CONVERSATION_STARTED = "conversation_started"
    CONVERSATION_TAG_ADDED = "conversation_tag_added"
    CONVERSATION_TAG_REMOVED = "conversation_tag_removed"
    MESSAGE_SENT = "message_sent"


class MessageMediaKind(str, Enum):
    NONE = "none"
    TEXT = "text"
    # images, videos, etc.


@dataclass(frozen=True)
class Event:
    """Describes arbitrary event within Suppgram application, with all relevant entities linked by their IDs."""

    kind: EventKind
    time_utc: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    agent_id: Optional[Any] = None
    conversation_id: Optional[Any] = None
    customer_id: Optional[Any] = None
    message_kind: Optional[MessageKind] = None
    message_media_kind: Optional[MessageMediaKind] = None
    tag_id: Optional[Any] = None
    workplace_id: Optional[Any] = None


# Data classes for observables


@dataclass(frozen=True)
class ConversationEvent:
    conversation: Conversation


@dataclass(frozen=True)
class ConversationTagEvent:
    conversation: Conversation
    tag: Tag


@dataclass(frozen=True)
class NewMessageForCustomerEvent:
    conversation: Conversation
    message: Message


@dataclass(frozen=True)
class NewUnassignedMessageFromCustomerEvent:
    message: Message
    conversation: Conversation


@dataclass(frozen=True)
class NewMessageForAgentEvent:
    agent: Agent
    workplace: Workplace
    message: Message


@dataclass(frozen=True)
class TagEvent:
    tag: Tag



from enum import Enum


class CallbackActionKind(str, Enum):
    # common
    PAGE = "page"

    # customer frontend
    RATE = "rate"

    # manager frontend
    ASSIGN_TO_ME = "assign_to_me"
    ADD_CONVERSATION_TAG = "add_tag"
    REMOVE_CONVERSATION_TAG = "remove_tag"



from dataclasses import replace
from typing import List, Optional, Any
from uuid import uuid4

from suppgram.frontends.telegram import TelegramStorage
from suppgram.frontends.telegram.storage import (
    TelegramMessage,
    TelegramMessageKind,
    TelegramChat,
    TelegramChatRole,
)


class InMemoryTelegramStorage(TelegramStorage):
    """In-memory implementation of [Storage][suppgram.storage.Storage] used in tests."""

    def __init__(self) -> None:
        self.chats: List[TelegramChat] = []
        self.messages: List[TelegramMessage] = []

    async def get_chat(self, telegram_chat_id: int) -> TelegramChat:
        try:
            return next(g for g in self.chats if g.telegram_chat_id == telegram_chat_id)
        except StopIteration:
            raise ValueError

    async def create_or_update_chat(self, telegram_chat_id: int) -> TelegramChat:
        try:
            return await self.get_chat(telegram_chat_id)
        except ValueError:
            chat = TelegramChat(telegram_chat_id=telegram_chat_id, roles=frozenset())
            self.chats.append(chat)
            return chat

    async def add_chat_roles(self, telegram_chat_id: int, *roles: TelegramChatRole):
        try:
            idx = next(
                i for i, g in enumerate(self.chats) if g.telegram_chat_id == telegram_chat_id
            )
        except StopIteration:
            raise ValueError
        chat = self.chats.pop(idx)
        chat = replace(chat, roles=chat.roles | {*roles})
        self.chats.append(chat)
        return chat

    async def get_chats_by_role(self, role: TelegramChatRole) -> List[TelegramChat]:
        return [g for g in self.chats if role in g.roles]

    async def insert_message(
        self,
        telegram_bot_id: int,
        chat: TelegramChat,
        telegram_message_id: int,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        customer_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None
    ) -> TelegramMessage:
        message = TelegramMessage(
            id=uuid4().hex,
            telegram_bot_id=telegram_bot_id,
            chat=chat,
            telegram_message_id=telegram_message_id,
            kind=kind,
            agent_id=agent_id,
            customer_id=customer_id,
            conversation_id=conversation_id,
            telegram_bot_username=telegram_bot_username,
        )
        self.messages.append(message)
        return message

    async def get_message(self, chat: TelegramChat, telegram_message_id: int) -> TelegramMessage:
        try:
            return next(
                m
                for m in self.messages
                if m.chat.telegram_chat_id == chat.telegram_chat_id
                and m.telegram_message_id == telegram_message_id
            )
        except StopIteration:
            raise ValueError

    async def get_messages(
        self,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None
    ) -> List[TelegramMessage]:
        return [
            m
            for m in self.messages
            if m.kind == kind
            and (agent_id is None or m.agent_id == agent_id)
            and (conversation_id is None or m.conversation_id == conversation_id)
            and (telegram_bot_username is None or m.telegram_bot_username == telegram_bot_username)
        ]

    async def delete_messages(self, messages: List[TelegramMessage]):
        message_ids = {m.id for m in messages}
        self.messages = [m for m in self.messages if m.id not in message_ids]

    async def get_newer_messages_of_kind(
        self, messages: List[TelegramMessage]
    ) -> List[TelegramMessage]:
        return [
            newer
            for newer in self.messages
            if any(
                newer.chat.telegram_chat_id == older.chat.telegram_chat_id
                and newer.telegram_message_id > older.telegram_message_id
                and newer.kind == older.kind
                for older in messages
            )
        ]



import html
import logging
import re
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Literal, Collection

from suppgram.emoji import EMOJI_SEQUENCE
from suppgram.entities import (
    Conversation,
    Customer,
    Tag,
    ConversationState,
    Agent,
    Message,
)
from suppgram.helpers import escape_markdown

logger = logging.getLogger(__name__)

TelegramParseMode = Optional[Literal["MarkdownV2", "HTML"]]


class Format(str, Enum):
    PLAIN = "plain"
    TELEGRAM_MARKDOWN = "telegram_markdown"
    TELEGRAM_HTML = "telegram_html"

    def convert_to_parse_mode(self) -> TelegramParseMode:
        if self == Format.PLAIN:
            return None
        if self == Format.TELEGRAM_MARKDOWN:
            return "MarkdownV2"
        if self == Format.TELEGRAM_HTML:
            return "HTML"
        raise ValueError(f"text format {self.value!r} is not directly supported by Telegram")

    @classmethod
    def get_formats_supported_by_telegram(cls) -> Collection["Format"]:
        return {cls.PLAIN, cls.TELEGRAM_MARKDOWN, cls.TELEGRAM_HTML}


@dataclass(frozen=True)
class Text:
    text: str
    format: Format = Format.PLAIN

    @property
    def parse_mode(self) -> Optional[str]:
        return self.format.convert_to_parse_mode()


class TextProvider:
    """Provides static texts and functions to compose dynamic texts where necessary."""

    telegram_customer_unsupported_message_content: str
    telegram_agent_unsupported_message_content: str

    telegram_customer_start_message: str
    telegram_customer_conversation_resolved_message_placeholder: str
    telegram_customer_conversation_resolved_message: str

    def compose_customer_conversation_resolved_message(self, rating: int) -> str:
        raise NotImplementedError

    telegram_manager_start_message: str
    telegram_manager_permission_denied_message: str
    telegram_agents_command_description: str
    telegram_agents_command_success_message: str
    telegram_send_new_conversations_command_description: str
    telegram_send_new_conversations_command_success_message: str

    telegram_report_command_description: str
    telegram_report_message_placeholder: str
    telegram_report_message: str

    # Tags
    telegram_create_tag_command_description: str
    telegram_create_tag_permission_denied_message: str
    telegram_create_tag_usage_message: str
    telegram_tag_already_exists_message: str
    telegram_tag_successfully_created_message: str

    def compose_add_tag_button_text(self, tag: Tag) -> str:
        return f"☐ {tag.name}"

    def compose_remove_tag_button_text(self, tag: Tag) -> str:
        return f"☑ {tag.name}"

    telegram_agent_start_message: str
    telegram_agent_permission_denied_message: str
    telegram_workplace_is_not_assigned_message: str
    telegram_resolve_command_description: str
    telegram_postpone_command_description: str
    telegram_agent_conversation_resolved_message: str

    telegram_new_conversation_notification_placeholder: str

    customer_profile_header: str
    customer_profile_anonymous: str
    customer_profile_contacts: str

    def compose_customer_profile(
        self, customer: Customer, allowed_formats: Collection[Format] = (Format.PLAIN,)
    ) -> Text:
        format_ = next(iter(allowed_formats))
        if Format.TELEGRAM_HTML in allowed_formats:
            format_ = Format.TELEGRAM_HTML
        elif Format.TELEGRAM_MARKDOWN in allowed_formats:
            format_ = Format.TELEGRAM_MARKDOWN

        full_name = (
            f"{customer.telegram_first_name or ''} {customer.telegram_last_name or ''}".strip()
            or self.customer_profile_anonymous
        )
        lines = [self.customer_profile_header.format(customer=full_name)]
        contacts = []
        if customer.telegram_user_id:
            contacts.append(
                self._format_telegram_mention(
                    telegram_user_id=customer.telegram_user_id,
                    telegram_first_name="Telegram",
                    telegram_last_name=None,
                    telegram_username=customer.telegram_username,
                    format_=format_,
                )
            )
        if contacts:
            lines.append(self.customer_profile_contacts.format(contacts=", ".join(contacts)))
        return Text(text="\n".join(lines), format=format_)

    CONVERSATION_STATE_TO_EMOJI = {
        ConversationState.NEW: "❗️",
        ConversationState.ASSIGNED: "⏳",
        ConversationState.RESOLVED: "✅",
    }

    def compose_conversation_notification_header(self, conversation: Conversation) -> str:
        raise NotImplementedError

    conversation_notification_assigned_to: str

    def compose_telegram_conversation_notification(self, conversation: Conversation) -> Text:
        profile = self.compose_customer_profile(
            conversation.customer,
            allowed_formats=Format.get_formats_supported_by_telegram(),
        )
        lines = [
            self.compose_conversation_notification_header(conversation),
            "",
            profile.text,
            "",
        ]
        lines.extend(self.format_history_message(message) for message in conversation.messages)
        if agent := conversation.assigned_agent:
            if agent.telegram_user_id:
                agent_ref = self._format_telegram_mention(
                    telegram_user_id=agent.telegram_user_id,
                    telegram_first_name=agent.telegram_first_name,
                    telegram_last_name=None,  # less formal
                    telegram_username=agent.telegram_username,
                    format_=profile.format,
                )
            else:
                logger.warning(f"Can't mention {agent} — unsupported agent frontend")
                agent_ref = f"#_{agent.id}"
            lines.extend(("", self.conversation_notification_assigned_to.format(agent=agent_ref)))
        if conversation.tags:
            tags = [self._format_telegram_tag(tag) for tag in conversation.tags]
            lines.extend(("", " ".join(tags)))
        return Text(text="\n".join(lines), format=profile.format)

    def compose_nudge_to_start_bot_notification(
        self, agent: Agent, telegram_bot_username: str
    ) -> Text:
        raise NotImplementedError

    telegram_assign_to_me_button_text: str

    def _format_telegram_mention(
        self,
        telegram_user_id: int,
        telegram_first_name: Optional[str],
        telegram_last_name: Optional[str],
        telegram_username: Optional[str],
        format_: Format,
    ) -> str:
        full_name: str = (
            f"{telegram_first_name or ''} {telegram_last_name or ''}".strip()
            or telegram_username
            or str(telegram_user_id)
        )

        if format_ == Format.PLAIN:
            if not telegram_username:
                logger.warning(
                    f"Can't mention Telegram user {telegram_user_id} without username in plain format"
                )
            return f"@{telegram_username}" if telegram_username else full_name

        url = f"tg://user?id={telegram_user_id}"

        if format_ == Format.TELEGRAM_MARKDOWN:
            escaped_name = escape_markdown(full_name)
            return f"[{escaped_name}]({url})"

        if format_ == Format.TELEGRAM_HTML:
            escaped_name = html.escape(full_name)
            return f'<a href="{url}">{escaped_name}</a>'

        raise ValueError(f"text format {format_.value!r} is not supported")

    _TAG_REGEX = re.compile(rf"^\s*({EMOJI_SEQUENCE}*)(.*?)({EMOJI_SEQUENCE}*)\s*$")

    def _format_telegram_tag(self, tag: Tag) -> str:
        match = self._TAG_REGEX.match(tag.name)
        if match is not None:
            prefix, tag_name, suffix = match.groups()
        else:
            logger.warning(f"Couldn't match tag {tag.name!r} against emoji regex")
            prefix, tag_name, suffix = "", tag.name, ""
        tag_name = re.sub(r"\s+", "_", tag_name)
        tag_name = re.sub(r"\W+", "", tag_name)
        tag_name = tag_name.strip("_")
        if not tag_name:
            tag_name = f"tag_{tag.id}"  # assuming that database IDs are safe
        return f"{prefix}#{tag_name}{suffix}"

    message_history_title: str

    def format_history_message(self, message: Message) -> str:
        raise NotImplementedError

    def format_rating(self, rating: int) -> str:
        return "★" * rating + "☆" * (5 - rating)
